# -*- coding: future_annotations -*-
from contextlib import contextmanager
import logging
from typing import TYPE_CHECKING

from nbsafety.safety import NotebookSafetySettings
from nbsafety.singletons import nbs
from test.utils import make_safety_fixture, skipif_known_failing

if TYPE_CHECKING:
    from typing import Dict

logging.basicConfig(level=logging.ERROR)

# Reset dependency graph before each test
# _safety_fixture, run_cell_ = make_safety_fixture(trace_messages_enabled=True)
_safety_fixture, run_cell_ = make_safety_fixture()


@contextmanager
def override_settings(**kwargs):
    old_settings = nbs().settings
    new_settings = old_settings._asdict()
    new_settings.update(kwargs)
    new_settings = NotebookSafetySettings(**new_settings)
    try:
        nbs().settings = new_settings
        yield
    finally:
        nbs().settings = old_settings


def run_cell(cell, cell_id, **kwargs):
    # print()
    # print('*******************************************')
    # print('running', cell)
    # print('*******************************************')
    # print()
    run_cell_(cell, cell_id=cell_id, **kwargs)


def run_all_cells(cells: Dict[int, str], **kwargs):
    for cell_id in sorted(cells.keys()):
        run_cell(cells[cell_id], cell_id=cell_id, **kwargs)


def test_simple():
    cells = {
        0: 'x = 0',
        1: 'y = x + 1',
        2: 'logging.info(y)',
        3: 'x = 42',
    }
    run_all_cells(cells)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == {2}
    assert response.fresh_cells == {1}
    assert response.stale_links == {2: {1}}
    assert response.refresher_links == {1: {2}}


def test_refresh_after_exception_fixed():
    cells = {
        0: 'x = 0',
        1: 'y = x + 1',
        2: 'logging.info(y)',
    }
    run_cell(cells[0], 0)
    run_cell(cells[2], 2, ignore_exceptions=True)
    run_cell(cells[1], 1)
    response = nbs().check_and_link_multiple_cells()
    assert response.fresh_cells == {2}


def test_refresh_after_val_changed():
    cells = {
        0: 'x = 0',
        1: 'y = x + 1',
        2: 'logging.info(y)',
        3: 'y = 42',
    }
    run_all_cells(cells)
    response = nbs().check_and_link_multiple_cells()
    assert response.fresh_cells == {2}


def test_inner_mutation_considered_fresh():
    cells = {
        0: 'lst_0 = [0,1,2]',
        1: 'lst_1 = [3,4,5]',
        2: 'lst = [lst_0, lst_1]',
        3: 'logging.info(lst)',
        4: 'lst_0.append(42)',
    }
    run_all_cells(cells)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == {2, 3}


# @pytest.mark.parametrize("force_subscript_symbol_creation", [True, False])
def test_update_list_elem():
    force_subscript_symbol_creation = True
    cells = {
        0: """
class Foo:
    def __init__(self):
        self.counter = 0
        self.dummy = 0
        
    def inc(self):
        self.counter += 1""",

        1: """
lst = []
for i in range(5):
    x = Foo()
    lst.append(x)""",

        2: """
for foo in lst:
    foo.inc()""",

        3: 'logging.info(lst)',
    }

    run_all_cells(cells)

    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == set(), 'got %s' % response.fresh_cells

    cells[4] = 'x.inc()'
    run_cell(cells[4], 4)

    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set(), 'got %s' % response.stale_cells
    assert response.fresh_cells == {2, 3}, 'got %s' % response.fresh_cells

    cells[5] = 'foo.inc()'
    run_cell(cells[5], 5)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set(), 'got %s' % response.stale_cells
    assert response.fresh_cells == {2, 3, 4}, 'got %s' % response.fresh_cells

    if force_subscript_symbol_creation:
        cells[6] = 'lst[-1]'
        run_cell(cells[6], 6)
        response = nbs().check_and_link_multiple_cells()
        assert response.stale_cells == set()
        assert response.fresh_cells == {2, 3, 4}

    run_cell(cells[4], 4)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == set([2, 3, 5] + ([6] if force_subscript_symbol_creation else []))


def test_no_freshness_for_alias_assignment_post_mutation():
    cells = {
        0: 'x = []',
        1: 'y = x',
        2: 'x.append(5)',
    }
    run_all_cells(cells)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == set()


def test_fresh_after_import():
    cells = {
        0: 'x = np.random.random(10)',
        1: 'import numpy as np'
    }
    run_all_cells(cells, ignore_exceptions=True)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == {0}


def test_external_object_update_propagates_to_stale_namespace_symbols():
    cells = {
        0: 'import fakelib',
        1: 'foo = fakelib.Foo()',
        2: 'logging.info(foo.x)',
        3: 'x = 42',
        4: 'foo.x = x + 1',
        5: 'x = 43',
        6: 'foo = foo.set_x(10)',
    }
    with override_settings(mark_stale_symbol_usages_unsafe=False):
        run_all_cells(cells)
        response = nbs().check_and_link_multiple_cells()
        assert response.stale_cells == set(), 'got %s' % response.stale_cells
        assert response.fresh_cells == {2, 4}


def test_symbol_on_both_sides_of_assignment():
    cells = {
        0: 'x = 0',
        1: 'y = x + 1',
        2: 'y += 7',
        3: 'x = 42',
    }
    run_all_cells(cells)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == {2}
    assert response.fresh_cells == {1}
    assert list(response.refresher_links.keys()) == [1]


def test_updated_namespace_after_subscript_dep_removed():
    cells = {
        0: 'x = 5',
        1: 'd = {x: 5}',
        2: 'logging.info(d[5])',
        3: 'x = 9',
    }
    run_all_cells(cells)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == {2}
    assert response.fresh_cells == {1}
    cells[1] = 'd = {5: 6}'
    run_cell(cells[1], 1)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == {2}
    run_cell(cells[2], 2)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == set()
    run_cell(cells[0], 0)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == set(), 'got %s' % response.fresh_cells


def test_equal_list_update_does_induce_fresh_cell():
    cells = {
        0: 'x = ["f"] + ["o"] * 10',
        1: 'y = x + list("bar")',
        2: 'logging.info(y)',
        3: 'y = list("".join(y))',
    }
    run_all_cells(cells)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == {2}
    run_cell('y = ("f",)', 4)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == {2, 3}


def test_equal_list_update_does_induce_fresh_cell_LITERAL_WITH_F_IS_REUSED_ON_UBUNTU_20_04_2_PYTHON_3_8_11():
    cells = {
        0: 'x = ["f"] + ["o"] * 10',
        1: 'y = x + list("bar")',
        2: 'logging.info(y)',
        3: 'y = list("".join(y))',
    }
    run_all_cells(cells)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == {2}
    run_cell('y = ["f"]', 4)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == {2, 3}


def test_equal_dict_update_does_induce_fresh_cell():
    cells = {
        0: 'x = {"foo": 42, "bar": 43}',
        1: 'y = dict(set(x.items()) | set({"baz": 44}.items()))',
        2: 'logging.info(y)',
        3: 'y = dict(y.items())',
    }
    run_all_cells(cells)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == {2}, 'got %s' % response.fresh_cells
    run_cell('y = {"foo": 99}', 4)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == {2, 3}


def test_list_append():
    cells = {
        0: 'lst = [0, 1]',
        1: 'x = lst[1] + 1',
        2: 'logging.info(x)',
        3: 'lst.append(2)',
    }
    run_all_cells(cells)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == set()
    run_cell('lst[1] += 42', 4)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == {2}
    assert response.fresh_cells == {1}


def test_list_extend():
    cells = {
        0: 'lst = [0, 1]',
        1: 'x = lst[1] + 1',
        2: 'logging.info(x)',
        3: 'lst.extend([2, 3, 4])',
    }
    run_all_cells(cells)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == set()
    run_cell('lst[1] += 42', 4)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == {2}
    assert response.fresh_cells == {1}


def test_implicit_subscript_symbol_does_not_bump_ts():
    cells = {
        0: 'lst = [] + [0, 1]',
        1: 'logging.info(lst)',
        2: 'logging.info(lst[0])',
    }
    run_all_cells(cells)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == set()


def test_liveness_skipped_for_simple_assignment_involving_aliases():
    cells = {
        0: 'lst = [1, 2, 3]',
        1: 'lst2 = lst',
        2: 'lst.append(4)',
    }
    run_all_cells(cells)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == set()
    run_cell('lst = [1, 2, 3, 4]', 3)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == {1, 2}, 'got %s' % response.fresh_cells


def test_incorrect_object_not_used_for_argument_symbols():
    cells = {
        0: 'import numpy as np',
        1: 'arr = np.random.randn(100)',
        2: 'def f(x): return []',
        # when obj for `np` was passed to `x`, this created issues
        3: 'f(np.arange(10))',
    }
    run_all_cells(cells)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == set(), 'got %s' % response.fresh_cells


def test_increment_by_same_amount():
    cells = {
        0: 'x = 2',
        1: 'y = x + 1',
        2: 'logging.info(y)',
    }
    run_all_cells(cells)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == set()
    run_cell('x = 3', 0)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == {2}
    assert response.fresh_cells == {1}


def test_list_insert():
    cells = {
        0: 'lst = [0, 1, 2, 4, 5, 6]',
        1: 'logging.info(lst[0])',
        2: 'logging.info(lst[1])',
        3: 'logging.info(lst[2])',
        4: 'logging.info(lst[3])',
        5: 'logging.info(lst[4])',
        6: 'logging.info(lst[5])',
        7: 'x = lst[5] + 42',
        8: 'logging.info(x)',
        9: 'lst.insert(3, 3)',
    }
    run_all_cells(cells)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == {8}, 'got %s' % response.stale_cells
    assert response.fresh_cells == {4, 5, 6, 7}, 'got %s' % response.fresh_cells


def _test_list_delete_helper(last_cell):
    cells = {
        0: 'lst = [0, 1, 2, 3, 3, 4, 5, 6]',
        1: 'logging.info(lst[0])',
        2: 'logging.info(lst[1])',
        3: 'logging.info(lst[2])',
        4: 'logging.info(lst[3])',
        5: 'logging.info(lst[4])',
        6: 'logging.info(lst[5])',
        7: 'logging.info(lst[6])',
        8: 'logging.info(lst[7])',
        9: 'x = lst[6] + 42',
        10: 'logging.info(x)',
        11: last_cell,
    }
    run_all_cells(cells)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == {10}
    # TODO: ideally we would detect that lst[3] is the same after deleting
    #  and not consider cell 4 to be fresh
    # TODO: ideally cell 8 would be considered unsafe since the last entry
    #  no longer exists
    assert response.fresh_cells == {4, 5, 6, 7, 9}, 'got %s' % response.fresh_cells


def test_list_delete():
    _test_list_delete_helper('del lst[3]')


def test_list_pop():
    _test_list_delete_helper('lst.pop(3)')


def test_list_remove():
    _test_list_delete_helper('lst.remove(3)')


def test_list_clear():
    cells = {
        0: 'lst = [0]',
        1: 'logging.info(lst[0])',
        2: 'logging.info(lst)',
    }
    run_all_cells(cells)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == set()
    run_cell('lst.clear()', 3)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == {2}


def test_dict_clear():
    cells = {
        0: 'd = {0: 0}',
        1: 'logging.info(d[0])',
        2: 'logging.info(d)',
    }
    run_all_cells(cells)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == set()
    run_cell('d.clear()', 3)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == {2}


def test_adhoc_pandas_series_update():
    cells = {
        0: 'import pandas as pd',
        1: 'df = pd.DataFrame({1: [2, 3], 4: [5, 7]})',
        2: 'df["foo"] = [8, 9]',
        3: 'df.foo.dropna(inplace=True)',
    }
    run_all_cells(cells)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == set()
    run_cell('df["foo"] = [8, 9]', 4)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == {3}



# -*- coding: future_annotations -*-
import asyncio
import inspect
import sys

from ipykernel.ipkernel import IPythonKernel

from nbsafety.version import __version__
from nbsafety.safety import NotebookSafety
from nbsafety.singletons import nbs


class SafeKernel(IPythonKernel):
    implementation = 'kernel'
    implementation_version = __version__

    def __init__(self, **kwargs):
        original_settrace = sys.settrace
        super().__init__(**kwargs)
        NotebookSafety.instance(use_comm=True, settrace=original_settrace)
        import nest_asyncio
        # ref: https://github.com/erdewit/nest_asyncio
        nest_asyncio.apply()

    def init_metadata(self, parent):
        """
        Don't actually change the metadata; we just want to get the cell id
        out of the execution request.
        """
        cell_id = parent.get('metadata', {}).get('cellId', None)
        if cell_id is not None:
            nbs().set_active_cell(cell_id, position_idx=None)
        return super().init_metadata(parent)

    if inspect.iscoroutinefunction(IPythonKernel.do_execute):
        async def do_execute(self, code, silent, store_history=False, user_expressions=None, allow_stdin=False):
            super_ = super()

            async def _run_cell_func(cell):
                return await super_.do_execute(cell, silent, store_history, user_expressions, allow_stdin)

            if silent:
                # then it's probably a control message; don't run through nbsafety
                return await _run_cell_func(code)
            else:
                return await nbs().safe_execute(code, True, _run_cell_func)
    else:
        def do_execute(self, code, silent, store_history=False, user_expressions=None, allow_stdin=False):
            super_ = super()

            def _run_cell_func(cell):
                return super_.do_execute(cell, silent, store_history, user_expressions, allow_stdin)
            return next(iter(asyncio.get_event_loop().run_until_complete(
                asyncio.wait([nbs().safe_execute(code, False, _run_cell_func)])
            )[0])).result()



# -*- coding: future_annotations -*-
import ast
import astunparse
import logging
import re
import shlex
import subprocess
from collections import defaultdict
from typing import TYPE_CHECKING, NamedTuple

from nbsafety.analysis.live_refs import (
    compute_live_dead_symbol_refs,
    get_symbols_for_references,
    get_live_symbols_and_cells_for_references,
)
from nbsafety.data_model.timestamp import Timestamp
from nbsafety.ipython_utils import cell_counter as ipy_cell_counter
from nbsafety.singletons import nbs

if TYPE_CHECKING:
    from typing import Dict, Generator, List, Optional, Set, Type, TypeVar
    from nbsafety.data_model.data_symbol import DataSymbol
    from nbsafety.types import CellId
    TimestampOrCounter = TypeVar('TimestampOrCounter', Timestamp, int)


logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)


_NB_MAGIC_PATTERN = re.compile(r'(^%|^!|^cd |\?$)')


class CheckerResult(NamedTuple):
    live: Set[DataSymbol]           # all live symbols in the cell
    deep_live: Set[DataSymbol]      # live symbols used in their entirety
    shallow_live: Set[DataSymbol]   # live symbols for which only a portion (attr or subscript) is used
    used_cells: Set[int]            # last updated timestamps of the live symbols
    live_cells: Set[int]            # cells that define a symbol that was called in the cell
    dead: Set[DataSymbol]           # symbols that are definitely assigned to
    stale: Set[DataSymbol]          # live symbols that have one or more ancestors with more recent timestamps
    typechecks: bool                # whether the cell typechecks successfully


def cells() -> Type[ExecutedCodeCell]:
    return ExecutedCodeCell


class ExecutedCodeCell:
    _current_cell_by_cell_id: Dict[CellId, ExecutedCodeCell] = {}
    _cell_by_cell_ctr: Dict[int, ExecutedCodeCell] = {}
    _cell_counter: int = 0
    _position_by_cell_id: Dict[CellId, int] = {}
    position_independent = True

    def __init__(self, cell_id: CellId, cell_ctr: int, content: str) -> None:
        self.cell_id: CellId = cell_id
        self.cell_ctr: int = cell_ctr
        self.content: str = content
        self._used_cell_counters_by_live_symbol: Dict[DataSymbol, Set[int]] = defaultdict(set)
        self._cached_ast: Optional[ast.Module] = None
        self._cached_typecheck_result: Optional[bool] = None if nbs().settings.mark_typecheck_failures_unsafe else True

    def __str__(self):
        return self.content

    def __repr__(self):
        return f'<{self.__class__.__name__}[id={self.cell_id},ctr={self.cell_ctr}]>'

    def __hash__(self):
        return hash((self.cell_id, self.cell_ctr))

    def add_used_cell_counter(self, sym: DataSymbol, ctr: int) -> None:
        if ctr > 0:
            self._used_cell_counters_by_live_symbol[sym].add(ctr)

    @classmethod
    def create_and_track(
        cls, cell_id: CellId, content: str, validate_ipython_counter: bool = True
    ) -> ExecutedCodeCell:
        cls._cell_counter += 1
        cell_ctr = cls._cell_counter
        if validate_ipython_counter:
            assert cell_ctr == ipy_cell_counter()
        cell = cls(cell_id, cell_ctr, content)
        cls._cell_by_cell_ctr[cell_ctr] = cell
        cur_cell = cls._current_cell_by_cell_id.get(cell_id, None)
        cur_cell_ctr = None if cur_cell is None else cur_cell.cell_ctr
        if cur_cell_ctr is None or cell_ctr > cur_cell_ctr:
            cls._current_cell_by_cell_id[cell_id] = cell
        return cell

    @classmethod
    def clear(cls):
        cls._current_cell_by_cell_id = {}
        cls._cell_by_cell_ctr = {}
        cls._cell_counter = 0
        cls._position_by_cell_id = {}

    @classmethod
    def set_cell_positions(cls, order_index_by_cell_id: Optional[Dict[CellId, int]]):
        if order_index_by_cell_id is None:
            cls.position_independent = True
            cls._position_by_cell_id = {}
        else:
            cls.position_independent = False
            cls._position_by_cell_id = order_index_by_cell_id

    @property
    def position(self) -> int:
        return self._position_by_cell_id.get(self.cell_id, -1)

    @classmethod
    def exec_counter(cls) -> int:
        return cls._cell_counter

    @classmethod
    def next_exec_counter(cls) -> int:
        return cls.exec_counter() + 1

    @classmethod
    def all_cells_most_recently_run_for_each_id(cls) -> Generator[ExecutedCodeCell, None, None]:
        yield from cls._current_cell_by_cell_id.values()

    @classmethod
    def from_timestamp(cls, ts: TimestampOrCounter) -> ExecutedCodeCell:
        if isinstance(ts, Timestamp):
            return cls._cell_by_cell_ctr[ts.cell_num]
        else:
            return cls._cell_by_cell_ctr[ts]

    @classmethod
    def from_id(cls, cell_id: CellId) -> Optional[ExecutedCodeCell]:
        return cls._current_cell_by_cell_id.get(cell_id, None)

    def sanitized_content(self):
        lines = []
        for line in self.content.strip().split('\n'):
            # TODO: figure out more robust strategy for filtering / transforming lines for the ast parser
            # we filter line magics, but for %time, we would ideally like to trace the statement being timed
            # TODO: how to do this?
            if _NB_MAGIC_PATTERN.search(line.strip()) is None:
                lines.append(line)
        return '\n'.join(lines)

    def to_ast(self, override: Optional[ast.Module] = None) -> ast.Module:
        if override is not None:
            self._cached_ast = override
        if self._cached_ast is None:
            self._cached_ast = ast.parse(self.sanitized_content())
        return self._cached_ast

    @property
    def is_current_for_id(self) -> bool:
        return self._current_cell_by_cell_id.get(self.cell_id, None) is self

    @classmethod
    def current_cell(cls) -> ExecutedCodeCell:
        return cls._cell_by_cell_ctr[cls._cell_counter]

    def get_max_used_live_symbol_cell_counter(self, live_symbols: Set[DataSymbol]) -> int:
        max_used_cell_ctr = -1
        this_cell_pos = self.position
        for sym in live_symbols:
            for cell_ctr in self._used_cell_counters_by_live_symbol.get(sym, []):
                if self.from_timestamp(cell_ctr).position <= this_cell_pos:
                    max_used_cell_ctr = max(max_used_cell_ctr, cell_ctr)
        return max_used_cell_ctr

    def check_and_resolve_symbols(
        self, update_liveness_time_versions: bool = False
    ) -> CheckerResult:
        live_symbol_refs, dead_symbol_refs = compute_live_dead_symbol_refs(self.to_ast(), scope=nbs().global_scope)
        if update_liveness_time_versions:
            get_live_symbols_and_cells_for_references(
                live_symbol_refs, nbs().global_scope, self.cell_ctr, update_liveness_time_versions=True
            )
        deep_live_symbols, shallow_live_symbols, live_cells = get_live_symbols_and_cells_for_references(
            live_symbol_refs, nbs().global_scope, self.cell_ctr, update_liveness_time_versions=False
        )
        live_symbols = deep_live_symbols | shallow_live_symbols
        # only mark dead attrsubs as killed if we can traverse the entire chain
        dead_symbols, _ = get_symbols_for_references(
            dead_symbol_refs, nbs().global_scope, only_yield_successful_resolutions=True
        )
        stale_symbols = {dsym for dsym in live_symbols if dsym.is_stale}
        for sym in deep_live_symbols:
            sym.cells_where_deep_live.add(self)
            self.add_used_cell_counter(sym, sym.timestamp.cell_num)
        for sym in shallow_live_symbols:
            sym.cells_where_shallow_live.add(self)
            self.add_used_cell_counter(sym, sym.timestamp_excluding_ns_descendents.cell_num)
        return CheckerResult(
            live=live_symbols,
            deep_live=deep_live_symbols,
            shallow_live=shallow_live_symbols,
            used_cells={
                sym.timestamp.cell_num for sym in deep_live_symbols
            } | {
                sym.timestamp_excluding_ns_descendents.cell_num for sym in shallow_live_symbols
            },
            live_cells=live_cells,
            dead=dead_symbols,
            stale=stale_symbols,
            typechecks=self._typechecks(live_cells, live_symbols),
        )

    def compute_phantom_cell_info(self, used_cells: Set[int]) -> Dict[CellId, Set[int]]:
        used_cell_counters_by_cell_id = defaultdict(set)
        used_cell_counters_by_cell_id[self.cell_id].add(self.exec_counter())
        for cell_num in used_cells:
            used_cell_counters_by_cell_id[self.from_timestamp(cell_num).cell_id].add(cell_num)
        return {
            cell_id: cell_execs
            for cell_id, cell_execs in used_cell_counters_by_cell_id.items()
            if len(cell_execs) >= 2
        }

    def _build_typecheck_slice(self, live_cell_ctrs: Set[int], live_symbols: Set[DataSymbol]) -> str:
        # TODO: typecheck statically-resolvable nested symbols too, not just top-level
        live_cell_counters = {self.cell_ctr}
        for live_cell_num in live_cell_ctrs:
            if self.from_timestamp(live_cell_num).is_current_for_id:
                live_cell_counters.add(live_cell_num)
        live_cells = [self.from_timestamp(ctr) for ctr in sorted(live_cell_counters)]
        top_level_symbols = {sym.get_top_level() for sym in live_symbols}
        top_level_symbols.discard(None)
        return '{type_declarations}\n\n{content}'.format(
            type_declarations='\n'.join(f'{sym.name}: {sym.get_type_annotation_string()}' for sym in top_level_symbols),
            content='\n'.join(live_cell.sanitized_content() for live_cell in live_cells),
        )

    def _typechecks(self, live_cell_ctrs: Set[int], live_symbols: Set[DataSymbol]) -> bool:
        if self._cached_typecheck_result is not None:
            return self._cached_typecheck_result
        typecheck_slice = self._build_typecheck_slice(live_cell_ctrs, live_symbols)
        try:
            # TODO: parse the output in order to pass up to the user
            ret = subprocess.call(f"mypy -c {shlex.quote(typecheck_slice)}", shell=True)
            self._cached_typecheck_result = (ret == 0)
        except Exception:
            logger.exception('Exception occurred during type checking')
            self._cached_typecheck_result = True
        return self._cached_typecheck_result

    @property
    def needs_typecheck(self):
        return self._cached_typecheck_result is None

    def invalidate_typecheck_result(self):
        self._cached_typecheck_result = None

    def compute_slice(self, stmt_level: bool = False) -> Dict[int, str]:
        """
        Gets a dictionary object of cell dependencies for the cell with
        the specified execution counter.

        Args:
            - cell_num (int): cell to get dependencies for, defaults to last
                execution counter

        Returns:
            - dict (int, str): map from required cell number to code
                representing dependencies
        """
        if stmt_level:
            stmts_by_cell_num = self.compute_slice_stmts()
            stmts_by_cell_num.pop(self.cell_ctr, None)
            ret = {
                ctr: '\n'.join(astunparse.unparse(stmt).strip() for stmt in stmts)
                for ctr, stmts in stmts_by_cell_num.items()
            }
            ret[self.cell_ctr] = self.content
            return ret
        else:
            deps: Set[int] = _compute_slice_impl(self.cell_ctr)
            return {dep: self.from_timestamp(dep).content for dep in deps}

    def compute_slice_stmts(self) -> Dict[int, List[ast.stmt]]:
        deps_stmt: Set[Timestamp] = _compute_slice_impl(Timestamp(self.cell_ctr, -1))
        stmts_by_cell_num = defaultdict(list)
        seen_stmt_ids = set()
        for ts in sorted(deps_stmt):
            if ts.cell_num > self.cell_ctr:
                break
            stmt = self.from_timestamp(ts.cell_num).to_ast().body[ts.stmt_num]
            stmt_id = id(stmt)
            if stmt is None or stmt_id in seen_stmt_ids:
                continue
            seen_stmt_ids.add(stmt_id)
            if stmt is not None:
                stmts_by_cell_num[ts.cell_num].append(stmt)
        stmts_by_cell_num[self.cell_ctr] = list(self.to_ast().body)
        return dict(stmts_by_cell_num)


def _compute_slice_impl(seed_ts: TimestampOrCounter) -> Set[TimestampOrCounter]:
    dependencies: Set[TimestampOrCounter] = set()
    timestamp_to_dynamic_ts_deps: Dict[TimestampOrCounter, Set[TimestampOrCounter]] = defaultdict(set)
    timestamp_to_static_ts_deps: Dict[TimestampOrCounter, Set[TimestampOrCounter]] = defaultdict(set)

    for sym in nbs().all_data_symbols():
        if nbs().mut_settings.dynamic_slicing_enabled:
            for used_time, sym_timestamp_when_used in sym.timestamp_by_used_time.items():
                if sym_timestamp_when_used < used_time:
                    if isinstance(seed_ts, Timestamp):
                        timestamp_to_dynamic_ts_deps[used_time].add(sym_timestamp_when_used)
                    else:
                        timestamp_to_dynamic_ts_deps[used_time.cell_num].add(sym_timestamp_when_used.cell_num)
        if nbs().mut_settings.static_slicing_enabled:
            for liveness_time, sym_timestamp_when_used in list(sym.timestamp_by_liveness_time.items()):
                if sym_timestamp_when_used < liveness_time:
                    if isinstance(seed_ts, Timestamp):
                        timestamp_to_static_ts_deps[liveness_time].add(sym_timestamp_when_used)
                    else:
                        timestamp_to_static_ts_deps[liveness_time.cell_num].add(
                            sym_timestamp_when_used.cell_num
                        )

    # ensure we at least get the static deps
    _get_ts_dependencies(
        seed_ts, dependencies, timestamp_to_dynamic_ts_deps, timestamp_to_static_ts_deps
    )
    if isinstance(seed_ts, Timestamp):
        for ts in list(timestamp_to_dynamic_ts_deps.keys() | timestamp_to_static_ts_deps.keys()):
            if ts.cell_num == seed_ts.cell_num:
                _get_ts_dependencies(
                    ts, dependencies, timestamp_to_dynamic_ts_deps, timestamp_to_static_ts_deps
                )
    return dependencies


def _get_ts_dependencies(
    timestamp: TimestampOrCounter,
    dependencies: Set[TimestampOrCounter],
    timestamp_to_dynamic_ts_deps: Dict[TimestampOrCounter, Set[TimestampOrCounter]],
    timestamp_to_static_ts_deps: Dict[TimestampOrCounter, Set[TimestampOrCounter]],
) -> None:
    """
    For a given timestamps, this function recursively populates a set of
    timestamps that the given timestamp depends on, based on the live symbols.

    Args:
        - timestamp (ts_or_int): current timestamp / cell to get dependencies for
        - dependencies (set<ts_or_int>): set of timestamps / cell coutners so far that exist
        - cell_num_to_dynamic_deps (dict<ts_or_int, set<ts_or_int>>): mapping from used timestamp
        to timestamp of symbol definition
        - timestamp_to_static_ts_deps (dict<ts_or_int, set<ts_or_int>>): mapping from used timestamp
        to timestamp of symbol definition (for statically computed timestamps)

    Returns:
        None
    """
    # Base case: cell already in dependencies
    if timestamp in dependencies:
        return
    if isinstance(timestamp, int) and timestamp < 0:
        return
    if isinstance(timestamp, Timestamp) and not timestamp.is_initialized:
        return

    # Add current cell to dependencies
    dependencies.add(timestamp)

    # Retrieve cell numbers for the dependent symbols
    # Add dynamic and static dependencies
    dep_timestamps = timestamp_to_dynamic_ts_deps[timestamp]
    logger.info('dynamic ts deps for %s: %s', timestamp, dep_timestamps)
    static_ts_deps = timestamp_to_static_ts_deps[timestamp]
    dep_timestamps |= static_ts_deps
    logger.info('static ts deps for %s: %s', timestamp, static_ts_deps)

    # For each dependent cell, recursively get their dependencies
    for ts in dep_timestamps - dependencies:
        _get_ts_dependencies(
            ts, dependencies, timestamp_to_dynamic_ts_deps, timestamp_to_static_ts_deps
        )



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Tuple, Union
    from nbsafety.analysis.attr_symbols import AttrSubSymbolChain
    CellId = Union[str, int]
    SymbolRef = Union[str, AttrSubSymbolChain]
    SupportedIndexType = Union[str, int, Tuple[Union[str, int], ...]]



# -*- coding: future_annotations -*-
import ast
import inspect
import logging
import symtable
from typing import cast, TYPE_CHECKING

from IPython import get_ipython

from nbsafety.analysis.attr_symbols import AttrSubSymbolChain, CallPoint
from nbsafety.data_model.data_symbol import DataSymbol, DataSymbolType
from nbsafety.utils.misc_utils import GetterFallback
from nbsafety.singletons import nbs, tracer

if TYPE_CHECKING:
    from typing import Any, Dict, Generator, Iterable, List, Mapping, Optional, Set, Tuple, Union
    from nbsafety.data_model.namespace import Namespace
    from nbsafety.types import SupportedIndexType


logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)


def _try_get_pandas():
    pandas = None
    try:
        import pandas
    except ImportError:
        pass
    return pandas


class Scope:
    GLOBAL_SCOPE_NAME = '<module>'

    def __init__(
        self,
        scope_name: str = GLOBAL_SCOPE_NAME,
        parent_scope: Optional[Scope] = None,
        symtab: Optional[symtable.SymbolTable] = None,
    ):
        self.scope_name = str(scope_name)
        self.parent_scope = parent_scope  # None iff this is the global scope
        self.symtab = symtab
        self._data_symbol_by_name: Dict[SupportedIndexType, DataSymbol] = {}

    def __hash__(self):
        return hash(self.full_path)

    def __str__(self):
        return str(self.full_path)

    def __repr__(self):
        return str(self)

    def data_symbol_by_name(self, is_subscript=False):
        if is_subscript:
            raise ValueError('Only namespace scopes carry subscripts')
        return self._data_symbol_by_name

    @property
    def non_namespace_parent_scope(self) -> Optional[Scope]:
        # a scope nested inside of a namespace scope does not have access
        # to unqualified members of the namespace scope
        if self.is_global:
            return None
        if self.parent_scope.is_namespace_scope:
            return self.parent_scope.non_namespace_parent_scope
        return self.parent_scope

    def make_child_scope(self, scope_name) -> Scope:
        symtab = tracer().cur_cell_symtab if self.is_global else self.symtab
        child_symtab = None
        if symtab is not None:
            try:
                sym = symtab.lookup(scope_name)
                if sym.is_namespace():
                    child_symtab = sym.get_namespace()
            except KeyError:
                pass
            except ValueError:
                pass
        return Scope(scope_name, parent_scope=self, symtab=child_symtab)

    def put(self, name: SupportedIndexType, val: DataSymbol) -> None:
        self._data_symbol_by_name[name] = val
        val.containing_scope = self

    def lookup_data_symbol_by_name_this_indentation(self, name: SupportedIndexType, **_: Any) -> Optional[DataSymbol]:
        return self._data_symbol_by_name.get(name, None)

    def all_data_symbols_this_indentation(self):
        return self._data_symbol_by_name.values()

    def lookup_data_symbol_by_name(self, name: SupportedIndexType, **kwargs: Any) -> Optional[DataSymbol]:
        ret = self.lookup_data_symbol_by_name_this_indentation(name, **kwargs)
        if ret is None and self.non_namespace_parent_scope is not None:
            ret = self.non_namespace_parent_scope.lookup_data_symbol_by_name(name, **kwargs)
        return ret

    @staticmethod
    def _get_name_to_obj_mapping(obj: Any, dsym: DataSymbol) -> Mapping[SupportedIndexType, Any]:
        if obj is None:
            return get_ipython().ns_table['user_global']
        elif dsym is not None and dsym.is_subscript:
            return obj
        else:
            try:
                pandas = _try_get_pandas()
                if (pandas is not None) and isinstance(obj, pandas.DataFrame):
                    # FIXME: hack to get it working w/ pandas, which doesn't play nicely w/ inspect.getmembers
                    name_to_obj = {}
                    for col in obj.columns:
                        try:
                            name_to_obj[col] = getattr(obj, col)
                        except:
                            continue
                else:
                    name_to_obj = GetterFallback([obj.__dict__, getattr(type(obj), '__dict__', {})])  # type: ignore
            except:  # noqa
                return dict(inspect.getmembers(obj))
        return name_to_obj

    def gen_data_symbols_for_attrsub_chain(
        self, chain: AttrSubSymbolChain
    ) -> Generator[Tuple[DataSymbol, Optional[Union[SupportedIndexType, CallPoint]], bool, bool], None, None]:
        """
        Generates progressive symbols appearing in an AttrSub chain until
        this can no longer be done semi-statically (e.g. because one of the
        chain members is a CallPoint).
        """
        cur_scope = self
        for i, name in enumerate(chain.symbols):
            is_last = i == len(chain.symbols) - 1
            if isinstance(name, CallPoint):
                next_dsym = cur_scope.lookup_data_symbol_by_name(name.symbol)
                if next_dsym is not None:
                    yield next_dsym, None if is_last else chain.symbols[i + 1], True, False
                break
            next_dsym = cur_scope.lookup_data_symbol_by_name(name)
            if next_dsym is None:
                break
            else:
                yield next_dsym, None if is_last else chain.symbols[i + 1], False, is_last
            cur_scope = next_dsym.namespace
            if cur_scope is None:
                break

    def get_most_specific_data_symbol_for_attrsub_chain(
        self, chain: AttrSubSymbolChain
    ) -> Optional[Tuple[DataSymbol, Optional[Union[SupportedIndexType, CallPoint]], bool, bool]]:
        """
        Get most specific DataSymbol for the whole chain (stops at first point it cannot find nested, e.g. a CallPoint).
        """
        ret = None
        for dsym, next_ref, is_called, success in self.gen_data_symbols_for_attrsub_chain(chain):
            ret = dsym, next_ref, is_called, success
        return ret

    @staticmethod
    def _resolve_symbol_type(
        overwrite: bool = True,
        is_subscript: bool = False,
        is_function_def: bool = False,
        is_import: bool = False,
        is_anonymous: bool = False,
        class_scope: Optional[Scope] = None,
    ):
        assert not (class_scope is not None and (is_function_def or is_import))
        if is_function_def:
            assert overwrite
            assert not is_subscript
            return DataSymbolType.FUNCTION
        elif is_import:
            assert overwrite
            assert not is_subscript
            return DataSymbolType.IMPORT
        elif class_scope is not None:
            assert overwrite
            assert not is_subscript
            return DataSymbolType.CLASS
        elif is_subscript:
            return DataSymbolType.SUBSCRIPT
        elif is_anonymous:
            return DataSymbolType.ANONYMOUS
        else:
            return DataSymbolType.DEFAULT

    def upsert_data_symbol_for_name(
        self,
        name: SupportedIndexType,
        obj: Any,
        deps: Iterable[DataSymbol],
        stmt_node: ast.AST,
        overwrite: bool = True,
        is_subscript: bool = False,
        is_function_def: bool = False,
        is_import: bool = False,
        is_anonymous: bool = False,
        class_scope: Optional[Scope] = None,
        symbol_type: Optional[DataSymbolType] = None,
        propagate: bool = True,
        implicit: bool = False,
    ) -> DataSymbol:
        symbol_type = symbol_type or self._resolve_symbol_type(
            overwrite=overwrite,
            is_subscript=is_subscript,
            is_function_def=is_function_def,
            is_import=is_import,
            is_anonymous=is_anonymous,
            class_scope=class_scope
        )
        deps = set(deps)  # make a copy since we mutate it (see below fixme)
        dsym, prev_dsym, prev_obj = self._upsert_data_symbol_for_name_inner(
            name,
            obj,
            deps,  # FIXME: this updates deps, which is super super hacky
            symbol_type,
            stmt_node,
            implicit=implicit,
        )
        dsym.update_deps(
            deps,
            prev_obj=prev_obj,
            overwrite=overwrite,
            propagate=propagate,
            refresh=not implicit,
        )
        return dsym

    def _upsert_data_symbol_for_name_inner(
        self,
        name: SupportedIndexType,
        obj: Any,
        deps: Set[DataSymbol],
        symbol_type: DataSymbolType,
        stmt_node: ast.AST,
        implicit: bool = False,
    ) -> Tuple[DataSymbol, Optional[DataSymbol], Optional[Any]]:
        prev_obj = None
        prev_dsym = self.lookup_data_symbol_by_name_this_indentation(
            name, is_subscript=symbol_type == DataSymbolType.SUBSCRIPT, skip_cloned_lookup=True,
        )
        if implicit and symbol_type != DataSymbolType.ANONYMOUS:
            assert prev_dsym is None, 'expected None, got %s' % prev_dsym
        if prev_dsym is not None:
            prev_obj = DataSymbol.NULL if prev_dsym.obj is None else prev_dsym.obj
            # TODO: handle case where new dsym is of different type
            if name in self.data_symbol_by_name(prev_dsym.is_subscript) and prev_dsym.symbol_type == symbol_type:
                prev_dsym.update_obj_ref(obj, refresh_cached=False)
                # old_dsym.update_type(symbol_type)
                # if we're updating a pre-existing one, it should not be an implicit upsert
                assert stmt_node is not None
                prev_dsym.update_stmt_node(stmt_node)
                return prev_dsym, prev_dsym, prev_obj
            else:
                # In this case, we are copying from a class and we need the dsym from which we are copying
                # as able to propagate to the new dsym.
                # Example:
                # class Foo:
                #     shared = 99
                # foo = Foo()
                # foo.shared = 42  # old_dsym refers to Foo.shared here
                # Earlier, we were explicitly adding Foo.shared as a dependency of foo.shared as follows:
                # deps.add(old_dsym)
                # But it turns out not to be necessary because foo depends on Foo, and changing Foo.shared will
                # propagate up the namespace hierarchy to Foo, which propagates to foo, which then propagates to
                # all of foo's namespace children (e.g. foo.shared).
                # This raises the question of whether we should draw the foo <-> Foo edge, since irrelevant namespace
                # children could then also be affected (e.g. some instance variable foo.x).
                # Perhaps a better strategy is to prevent propagation along this edge unless class Foo is redeclared.
                # If we do this, then we should go back to explicitly adding the dep as follows:
                # EDIT: added check to avoid propagating along class -> instance edge when class not redefined, so now
                # it is important to explicitly add this dep.
                deps.add(prev_dsym)
        ns_self = self.namespace
        if ns_self is not None and symbol_type == DataSymbolType.DEFAULT and ns_self.cloned_from is not None:
            # add the cloned symbol as a dependency of the symbol about to be created
            new_dep = ns_self.cloned_from.lookup_data_symbol_by_name_this_indentation(name, is_subscript=False)
            if new_dep is not None:
                deps.add(new_dep)
        dsym = DataSymbol(
            name, symbol_type, obj, self, stmt_node=stmt_node, refresh_cached_obj=False, implicit=implicit
        )
        self.put(name, dsym)
        return dsym, prev_dsym, prev_obj

    def delete_data_symbol_for_name(self, name: SupportedIndexType, is_subscript: bool = False):
        assert not is_subscript
        dsym = self._data_symbol_by_name.pop(name, None)
        if dsym is not None:
            dsym.update_deps(set(), deleted=True)

    @property
    def is_global(self):
        return self.parent_scope is None

    @property
    def is_globally_accessible(self):
        return self.is_global or (self.is_namespace_scope and self.parent_scope.is_globally_accessible)

    @property
    def is_namespace_scope(self):
        return False

    @property
    def namespace(self) -> Optional[Namespace]:
        if self.is_namespace_scope:
            return cast('Namespace', self)
        else:
            return None

    @property
    def global_scope(self):
        if self.is_global:
            return self
        return self.parent_scope.global_scope

    @property
    def full_path(self) -> Tuple[str, ...]:
        path = (self.scope_name,)
        if self.is_global:
            return path
        else:
            return self.parent_scope.full_path + path

    @property
    def full_namespace_path(self) -> str:
        if not self.is_namespace_scope:
            return ''
        if self.parent_scope is not None:
            prefix = self.parent_scope.full_namespace_path
        else:
            prefix = ''
        if prefix:
            if getattr(self, 'is_subscript', False):
                return f'{prefix}[{self.scope_name}]'
            else:
                return f'{prefix}.{self.scope_name}'
        else:
            return self.scope_name

    def make_namespace_qualified_name(self, dsym: DataSymbol) -> str:
        return str(dsym.name)



# -*- coding: future_annotations -*-
import ast
from contextlib import contextmanager
from typing import Sequence, TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Union


class SkipUnboundArgsMixin(ast.NodeVisitor):
    # Only need to check for default arguments
    def visit_arguments(self, node):
        self.visit(node.defaults)
        self.visit(node.kw_defaults)


class VisitListsMixin(ast.NodeVisitor):
    def generic_visit(self, node: Union[ast.AST, Sequence[ast.AST]]):
        if node is None:
            return
        elif isinstance(node, Sequence):
            for item in node:
                self.visit(item)
        else:
            super().generic_visit(node)


class SaveOffAttributesMixin:
    @contextmanager
    def push_attributes(self, **kwargs):
        for k in kwargs:
            if not hasattr(self, k):
                raise AttributeError('requested to save unfound attribute %s of object %s' % (k, self))
        saved_attributes = {}
        for k in kwargs:
            saved_attributes[k] = getattr(self, k)
        for k, v in kwargs.items():
            setattr(self, k, v)
        yield
        for k, v in saved_attributes.items():
            setattr(self, k, v)



# -*- coding: future_annotations -*-
import ast
from typing import TYPE_CHECKING

from nbsafety.tracing.stmt_mapper import StatementMapper

if TYPE_CHECKING:
    from typing import Dict


def compute_lineno_to_stmt_mapping(code: str) -> Dict[int, ast.stmt]:
    mapper = StatementMapper(0, {}, {}, {})
    mapper(ast.parse(code))
    return mapper.line_to_stmt_map


def test_for_loop():
    code = """
for i in range(10):
    a: int = i
    b = a + i
    lst: List[int] = [a, b]
""".strip()
    mapping = compute_lineno_to_stmt_mapping(code)
    assert isinstance(mapping[1], ast.For)
    assert isinstance(mapping[2], ast.AnnAssign)
    assert isinstance(mapping[3], ast.Assign)
    assert isinstance(mapping[4], ast.AnnAssign)


def test_multiline_for_loop():
    code = """
for i in [
    0,
    1,
    2,
    3,
    4,
]:
    a = i
    b = a + i
    lst = [a, b]
""".strip()
    mapping = compute_lineno_to_stmt_mapping(code)
    # for i in range(1, 7):
    #     assert isinstance(mapping[i], ast.For)
    # assert 7 not in mapping
    assert isinstance(mapping[1], ast.For)
    for i in range(2, 8):
        assert i not in mapping
    assert isinstance(mapping[8], ast.Assign)
    assert isinstance(mapping[9], ast.Assign)
    assert isinstance(mapping[10], ast.Assign)


def test_if():
    code = """
if True:
    x = 0
else:
    x: int = 0
""".strip()
    mapping = compute_lineno_to_stmt_mapping(code)
    assert isinstance(mapping[1], ast.If)
    assert isinstance(mapping[2], ast.Assign)
    assert 3 not in mapping
    assert isinstance(mapping[4], ast.AnnAssign)



# -*- coding: future_annotations -*-
import ast
import asyncio
from collections import defaultdict
from contextlib import contextmanager
from dataclasses import dataclass
import inspect
import json
import logging
import re
import sys
from typing import cast, TYPE_CHECKING, NamedTuple

from IPython import get_ipython
from IPython.core.magic import register_cell_magic, register_line_magic

from nbsafety.ipython_utils import (
    ast_transformer_context,
    run_cell,
    save_number_of_currently_executing_cell,
)
from nbsafety import line_magics
from nbsafety.data_model.code_cell import cells, ExecutedCodeCell
from nbsafety.data_model.data_symbol import DataSymbol
from nbsafety.data_model.namespace import Namespace
from nbsafety.data_model.scope import Scope
from nbsafety.data_model.timestamp import Timestamp
from nbsafety.run_mode import ExecutionMode, SafetyRunMode
from nbsafety import singletons
from nbsafety.tracing.safety_ast_rewriter import SafetyAstRewriter
from nbsafety.tracing.trace_manager import TraceManager

if TYPE_CHECKING:
    from typing import Any, Dict, Iterable, List, Set, Optional, Tuple, Union
    from types import FrameType
    from nbsafety.types import CellId, SupportedIndexType

logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)

_SAFETY_LINE_MAGIC = 'safety'
_NB_MAGIC_PATTERN = re.compile(r'(^%|^!|^cd |\?$)')


class NotebookSafetySettings(NamedTuple):
    store_history: bool
    test_context: bool
    use_comm: bool
    backwards_cell_staleness_propagation: bool
    track_dependencies: bool
    mark_stale_symbol_usages_unsafe: bool
    mark_typecheck_failures_unsafe: bool
    mark_phantom_cell_usages_unsafe: bool
    mode: SafetyRunMode


@dataclass
class MutableNotebookSafetySettings:
    trace_messages_enabled: bool
    highlights_enabled: bool
    static_slicing_enabled: bool
    dynamic_slicing_enabled: bool
    exec_mode: ExecutionMode


class FrontendCheckerResult(NamedTuple):
    stale_cells: Set[CellId]
    fresh_cells: Set[CellId]
    new_fresh_cells: Set[CellId]
    stale_links: Dict[CellId, Set[CellId]]
    refresher_links: Dict[CellId, Set[CellId]]
    phantom_cell_info: Dict[CellId, Dict[CellId, Set[int]]]

    def to_json(self) -> Dict[str, Any]:
        return {
            'stale_cells': list(self.stale_cells),
            'fresh_cells': list(self.fresh_cells),
            'new_fresh_cells': list(self.new_fresh_cells),
            'stale_links': {cell_id: list(linked_cell_ids) for cell_id, linked_cell_ids in self.stale_links.items()},
            'refresher_links': {
                cell_id: list(linked_cell_ids) for cell_id, linked_cell_ids in self.refresher_links.items()
            },
        }


class NotebookSafety(singletons.NotebookSafety):
    """Holds all the state necessary to detect stale dependencies in Jupyter notebooks."""

    def __init__(self, cell_magic_name=None, use_comm=False, settrace=None, **kwargs):
        super().__init__()
        cells().clear()
        self.settings: NotebookSafetySettings = NotebookSafetySettings(
            store_history=kwargs.pop('store_history', True),
            test_context=kwargs.pop('test_context', False),
            use_comm=use_comm,
            backwards_cell_staleness_propagation=True,
            track_dependencies=True,
            mark_stale_symbol_usages_unsafe=kwargs.pop('mark_stale_symbol_usages_unsafe', True),
            mark_typecheck_failures_unsafe=kwargs.pop('mark_typecheck_failures_unsafe', False),
            mark_phantom_cell_usages_unsafe=kwargs.pop('mark_phantom_cell_usages_unsafe', False),
            mode=SafetyRunMode.get(),
        )
        self.mut_settings: MutableNotebookSafetySettings = MutableNotebookSafetySettings(
            trace_messages_enabled=kwargs.pop('trace_messages_enabled', False),
            highlights_enabled=kwargs.pop('highlights_enabled', True),
            static_slicing_enabled=kwargs.pop('static_slicing_enabled', True),
            dynamic_slicing_enabled=kwargs.pop('dynamic_slicing_enabled', True),
            exec_mode=ExecutionMode(kwargs.pop('exec_mode', ExecutionMode.NORMAL)),
        )
        # Note: explicitly adding the types helps PyCharm intellisense
        self.settrace = settrace or sys.settrace
        self.namespaces: Dict[int, Namespace] = {}
        self.aliases: Dict[int, Set[DataSymbol]] = defaultdict(set)
        self.global_scope: Scope = Scope()
        self.updated_symbols: Set[DataSymbol] = set()
        self.statement_cache: Dict[int, Dict[int, ast.stmt]] = defaultdict(dict)
        self.ast_node_by_id: Dict[int, ast.AST] = {}
        self.loop_iter_flag_names: Set[str] = set()
        self.parent_node_by_id: Dict[int, ast.AST] = {}
        self.statement_to_func_cell: Dict[int, DataSymbol] = {}
        self._active_cell_id: Optional[str] = None
        self.active_cell_position_idx = -1
        self.safety_issue_detected = False
        if cell_magic_name is None:
            self._cell_magic = None
        else:
            self._cell_magic = self._make_cell_magic(cell_magic_name)
        self._line_magic = self._make_line_magic()
        self._prev_cell_stale_symbols: Set[DataSymbol] = set()
        self._cell_name_to_cell_num_mapping: Dict[str, int] = {}
        self._exception_raised_during_execution: Optional[Exception] = None
        self._saved_debug_message: Optional[str] = None
        if use_comm:
            get_ipython().kernel.comm_manager.register_target(__package__, self._comm_target)

    @property
    def is_develop(self) -> bool:
        return self.settings.mode == SafetyRunMode.DEVELOP

    @property
    def is_test(self) -> bool:
        return self.settings.test_context

    @property
    def trace_messages_enabled(self) -> bool:
        return self.mut_settings.trace_messages_enabled

    @trace_messages_enabled.setter
    def trace_messages_enabled(self, new_val) -> None:
        self.mut_settings.trace_messages_enabled = new_val

    def get_first_full_symbol(self, obj_id: int) -> Optional[DataSymbol]:
        # TODO: also avoid anonymous namespaces?
        for alias in self.aliases.get(obj_id, []):
            if not alias.is_anonymous:
                return alias
        return None

    def cell_counter(self) -> int:
        return cells().exec_counter()

    def reset_cell_counter(self):
        # only called in test context
        assert not self.settings.store_history
        for sym in self.all_data_symbols():
            sym._timestamp = sym._max_inner_timestamp = sym.required_timestamp = Timestamp.uninitialized()
            sym.timestamp_by_used_time.clear()
            sym.timestamp_by_liveness_time.clear()
        cells().clear()

    def set_exception_raised_during_execution(self, new_val: Optional[Exception] = None) -> Optional[Exception]:
        ret = self._exception_raised_during_execution
        self._exception_raised_during_execution = new_val
        return ret

    def get_position(self, frame: FrameType) -> Tuple[Optional[int], int]:
        try:
            cell_num = self._cell_name_to_cell_num_mapping.get(frame.f_code.co_filename, None)
            return cell_num, frame.f_lineno
        except KeyError as e:
            logger.error('key error while retrieving cell for %s', frame.f_code.co_filename)
            raise e

    def set_name_to_cell_num_mapping(self, frame: FrameType):
        self._cell_name_to_cell_num_mapping[frame.f_code.co_filename] = cells().exec_counter()

    def is_cell_file(self, fname: str) -> bool:
        return fname in self._cell_name_to_cell_num_mapping

    def set_active_cell(self, cell_id, position_idx=-1):
        self._active_cell_id = cell_id
        if position_idx is not None:
            self.active_cell_position_idx = position_idx

    def _comm_target(self, comm, open_msg):
        @comm.on_msg
        def _responder(msg):
            request = msg['content']['data']
            self.handle(request, comm=comm)

        comm.send({'type': 'establish'})

    def handle(self, request, comm=None):
        if request['type'] == 'change_active_cell':
            self.set_active_cell(request['active_cell_id'], position_idx=request.get('active_cell_order_idx', -1))
        elif request['type'] == 'cell_freshness':
            if self._active_cell_id is None:
                self._active_cell_id = request.get('executed_cell_id', None)
            cell_id = request.get('executed_cell_id', None)
            if self.settings.backwards_cell_staleness_propagation:
                order_index_by_id = None
                cells_to_check = cells().all_cells_most_recently_run_for_each_id()
            else:
                order_index_by_id = request['order_index_by_cell_id']
                cells_to_check = (
                    cell for cell in (
                        cells().from_id(cell_id) for cell_id in order_index_by_id
                    ) if cell is not None
                )
            cells().set_cell_positions(order_index_by_id)
            response = self.check_and_link_multiple_cells(cells_to_check=cells_to_check).to_json()
            response['type'] = 'cell_freshness'
            response['exec_mode'] = self.mut_settings.exec_mode.value
            response['last_executed_cell_id'] = cell_id
            response['highlights_enabled'] = self.mut_settings.highlights_enabled
            if comm is not None:
                comm.send(response)
        else:
            dbg_msg = 'Unsupported request type for request %s' % request
            logger.error(dbg_msg)
            self._saved_debug_message = dbg_msg

    def check_and_link_multiple_cells(
        self,
        cells_to_check: Optional[Iterable[ExecutedCodeCell]] = None,
        update_liveness_time_versions: bool = False,
    ) -> FrontendCheckerResult:
        stale_cells = set()
        typecheck_error_cells = set()
        fresh_cells = set()
        new_fresh_cells = set()
        stale_symbols_by_cell_id: Dict[CellId, Set[DataSymbol]] = {}
        killing_cell_ids_for_symbol: Dict[DataSymbol, Set[CellId]] = defaultdict(set)
        phantom_cell_info: Dict[CellId, Dict[CellId, Set[int]]] = {}
        if cells_to_check is None:
            cells_to_check = cells().all_cells_most_recently_run_for_each_id()
        for cell in cells_to_check:
            try:
                checker_result = cell.check_and_resolve_symbols(
                    update_liveness_time_versions=update_liveness_time_versions
                )
            except SyntaxError:
                continue
            cell_id = cell.cell_id
            if cells().position_independent:
                stale_symbols = {sym for sym in checker_result.live if sym.is_stale}
            else:
                stale_symbols = {sym for sym in checker_result.live if sym.is_stale_at_position(cell.position)}
            if len(stale_symbols) > 0:
                stale_symbols_by_cell_id[cell_id] = stale_symbols
                stale_cells.add(cell_id)
            if not checker_result.typechecks:
                typecheck_error_cells.add(cell_id)
            for dead_sym in checker_result.dead:
                killing_cell_ids_for_symbol[dead_sym].add(cell_id)

            if self.settings.mark_phantom_cell_usages_unsafe:
                phantom_cell_info_for_cell = cell.compute_phantom_cell_info(checker_result.used_cells)
                if len(phantom_cell_info_for_cell) > 0:
                    phantom_cell_info[cell_id] = phantom_cell_info_for_cell

            if cell_id not in stale_cells:
                max_timestamp_cell_num = cell.get_max_used_live_symbol_cell_counter(checker_result.live)
                if max_timestamp_cell_num > cell.cell_ctr:
                    fresh_cells.add(cell_id)
                if max_timestamp_cell_num >= cells().exec_counter():
                    new_fresh_cells.add(cell_id)
        stale_links: Dict[CellId, Set[CellId]] = defaultdict(set)
        refresher_links: Dict[CellId, Set[CellId]] = defaultdict(set)
        for stale_cell_id in stale_cells:
            stale_syms = stale_symbols_by_cell_id[stale_cell_id]
            refresher_cell_ids: Set[CellId] = set()
            refresher_cell_ids = refresher_cell_ids.union(
                *(killing_cell_ids_for_symbol[stale_sym] for stale_sym in stale_syms)
            )
            stale_links[stale_cell_id] = refresher_cell_ids
        stale_link_changes = True
        # transitive closer up until we hit non-stale refresher cells
        while stale_link_changes:
            stale_link_changes = False
            for stale_cell_id in stale_cells:
                new_stale_links = set(stale_links[stale_cell_id])
                original_length = len(new_stale_links)
                for refresher_cell_id in stale_links[stale_cell_id]:
                    if refresher_cell_id not in stale_cells:
                        continue
                    new_stale_links |= stale_links[refresher_cell_id]
                new_stale_links.discard(stale_cell_id)
                stale_link_changes = stale_link_changes or original_length != len(new_stale_links)
                stale_links[stale_cell_id] = new_stale_links
        for stale_cell_id in stale_cells:
            stale_links[stale_cell_id] -= stale_cells
            for refresher_cell_id in stale_links[stale_cell_id]:
                refresher_links[refresher_cell_id].add(stale_cell_id)
        return FrontendCheckerResult(
            # TODO: we should probably have separate fields for stale vs non-typechecking cells,
            #  or at least change the name to a more general "unsafe_cells" or equivalent
            stale_cells=stale_cells | typecheck_error_cells,
            fresh_cells=fresh_cells,
            new_fresh_cells=new_fresh_cells,
            stale_links=stale_links,
            refresher_links=refresher_links,
            phantom_cell_info=phantom_cell_info,
        )

    @staticmethod
    def _get_max_timestamp_cell_num_for_symbols(deep_symbols: Set[DataSymbol], shallow_symbols: Set[DataSymbol]) -> int:
        max_timestamp_cell_num = -1
        for dsym in deep_symbols:
            max_timestamp_cell_num = max(max_timestamp_cell_num, dsym.timestamp.cell_num)
        for dsym in shallow_symbols:
            max_timestamp_cell_num = max(max_timestamp_cell_num, dsym.timestamp_excluding_ns_descendents.cell_num)
        return max_timestamp_cell_num

    def _safety_precheck_cell(self, cell: ExecutedCodeCell) -> None:
        checker_result = self.check_and_link_multiple_cells(
            cells_to_check=[cell],
            update_liveness_time_versions=self.mut_settings.static_slicing_enabled,
        )
        if cell.cell_id in checker_result.stale_cells:
            self.safety_issue_detected = True

    def _resync_symbols(self, symbols: Iterable[DataSymbol]):
        for dsym in symbols:
            if not dsym.containing_scope.is_global:
                continue
            obj = get_ipython().user_global_ns.get(dsym.name, None)
            if obj is None:
                continue
            if dsym.obj_id == id(obj):
                continue
            for alias in self.aliases[dsym.cached_obj_id] | self.aliases[dsym.obj_id]:
                containing_namespace = alias.containing_namespace
                if containing_namespace is None:
                    continue
                containing_obj = containing_namespace.obj
                if containing_obj is None:
                    continue
                # TODO: handle dict case too
                if isinstance(containing_obj, list) and containing_obj[-1] is obj:
                    containing_namespace._subscript_data_symbol_by_name.pop(alias.name, None)
                    alias.name = len(containing_obj) - 1
                    alias.update_obj_ref(obj)
                    containing_namespace._subscript_data_symbol_by_name[alias.name] = alias
            self.aliases[dsym.cached_obj_id].discard(dsym)
            self.aliases[dsym.obj_id].discard(dsym)
            self.aliases[id(obj)].add(dsym)
            dsym.update_obj_ref(obj)

    def create_dag_metadata(self) -> Dict[int, Dict[str, Union[List[int], List[str], Dict[str, Dict[str, str]]]]]:
        cell_num_to_used_imports: Dict[int, Set[DataSymbol]] = defaultdict(set)
        cell_num_to_dynamic_inputs: Dict[int, Set[DataSymbol]] = defaultdict(set)
        cell_num_to_dynamic_outputs: Dict[int, Set[DataSymbol]] = defaultdict(set)
        cell_num_to_dynamic_cell_parents: Dict[int, Set[int]] = defaultdict(set)
        cell_num_to_dynamic_cell_children: Dict[int, Set[int]] = defaultdict(set)

        for sym in self.all_data_symbols():
            top_level_sym = sym.get_top_level()
            if top_level_sym is None or not top_level_sym.is_globally_accessible or top_level_sym.is_anonymous:
                # TODO: also skip lambdas
                continue
            for used_time, sym_timestamp_when_used in sym.timestamp_by_used_time.items():
                if top_level_sym.is_import:
                    cell_num_to_used_imports[used_time.cell_num].add(top_level_sym)
                else:
                    if sym_timestamp_when_used < used_time:
                        cell_num_to_dynamic_cell_parents[used_time.cell_num].add(sym_timestamp_when_used.cell_num)
                        cell_num_to_dynamic_inputs[used_time.cell_num].add(top_level_sym)
                        cell_num_to_dynamic_cell_children[sym_timestamp_when_used.cell_num].add(used_time.cell_num)
                    cell_num_to_dynamic_outputs[sym_timestamp_when_used.cell_num].add(top_level_sym)
            if not top_level_sym.is_import:
                for updated_time in sym.updated_timestamps:
                    # TODO: distinguished between used / unused outputs?
                    cell_num_to_dynamic_outputs[updated_time.cell_num].add(top_level_sym)

        cell_metadata: Dict[int, Dict[str, Union[List[int], List[str], Dict[str, Dict[str, str]]]]] = {}
        all_relevant_cells = (
            cell_num_to_used_imports.keys() |
            cell_num_to_dynamic_inputs.keys() |
            cell_num_to_dynamic_outputs.keys() |
            cell_num_to_dynamic_cell_parents.keys() |
            cell_num_to_dynamic_cell_children.keys()
        )
        for cell_num in all_relevant_cells:
            cell_imports = [dsym.get_import_string() for dsym in cell_num_to_used_imports[cell_num]]
            input_symbols = {
                str(dsym): {
                    'type': dsym.get_type_annotation_string()
                } for dsym in cell_num_to_dynamic_inputs[cell_num]
            }
            output_symbols = {
                str(dsym): {
                    'type': dsym.get_type_annotation_string()
                } for dsym in cell_num_to_dynamic_outputs[cell_num]
            }
            parent_cells = list(cell_num_to_dynamic_cell_parents[cell_num])
            child_cells = list(cell_num_to_dynamic_cell_children[cell_num])
            cell_metadata[cell_num] = {
                'cell_imports': cell_imports,
                'input_symbols': input_symbols,
                'output_symbols': output_symbols,
                'parent_cells': parent_cells,
                'child_cells': child_cells,
            }
        return cell_metadata

    async def safe_execute(self, cell_content: str, is_async: bool, run_cell_func):
        if self._saved_debug_message is not None:  # pragma: no cover
            logger.error(self._saved_debug_message)
            self._saved_debug_message = None
        ret = None
        with save_number_of_currently_executing_cell():
            cell_id, self._active_cell_id = self._active_cell_id, None
            assert cell_id is not None
            cell = cells().create_and_track(
                cell_id, cell_content, validate_ipython_counter=self.settings.store_history
            )

            # Stage 1: Precheck.
            self._safety_precheck_cell(cell)

            # Stage 2: Trace / run the cell, updating dependencies as they are encountered.
            try:
                with self._tracing_context(cell_id):
                    if is_async:
                        ret = await run_cell_func(cell_content)  # pragma: no cover
                    else:
                        ret = run_cell_func(cell_content)
                # Stage 2.1: resync any defined symbols that could have gotten out-of-sync
                #  due to tracing being disabled

                self._resync_symbols([
                    # TODO: avoid bad performance by only iterating over symbols updated in this cell
                    sym for sym in self.all_data_symbols() if sym.timestamp.cell_num == cells().exec_counter()
                ])
                self._gc()
            except Exception as e:
                if self.is_test:
                    self.set_exception_raised_during_execution(e)
            finally:
                return ret

    def _make_cell_magic(self, cell_magic_name):
        # this is to avoid capturing `self` and creating an extra reference to the singleton
        store_history = self.settings.store_history

        def _run_cell_func(cell):
            run_cell(cell, store_history=store_history)

        def _dependency_safety(_, cell: str):
            asyncio.get_event_loop().run_until_complete(
                asyncio.wait([singletons.nbs().safe_execute(cell, False, _run_cell_func)])
            )

        # FIXME (smacke): probably not a great idea to rely on this
        _dependency_safety.__name__ = cell_magic_name
        return register_cell_magic(_dependency_safety)

    @contextmanager
    def _tracing_context(self, cell_id: CellId):
        self.updated_symbols.clear()

        try:
            with TraceManager.instance().tracing_context():
                with ast_transformer_context([SafetyAstRewriter(cell_id)]):
                    yield
        finally:
            TraceManager.clear_instance()

    def _make_line_magic(self):
        print_ = print  # to keep the test from failing since this is a legitimate print
        line_magic_names = [f[0] for f in inspect.getmembers(line_magics) if inspect.isfunction(f[1])]

        def _handle(cmd, line):
            if cmd in ('deps', 'show_deps', 'show_dependency', 'show_dependencies'):
                return line_magics.show_deps(line)
            elif cmd in ('stale', 'show_stale'):
                return line_magics.show_stale(line)
            elif cmd == 'trace_messages':
                return line_magics.trace_messages(line)
            elif cmd in ('hls', 'nohls', 'highlight', 'highlights'):
                return line_magics.set_highlights(cmd, line)
            elif cmd in ('dag', 'make_dag', 'cell_dag', 'make_cell_dag'):
                return json.dumps(self.create_dag_metadata(), indent=2)
            elif cmd in ('slice', 'make_slice', 'gather_slice'):
                return line_magics.make_slice(line)
            elif cmd == 'remove_dependency':
                return line_magics.remove_dep(line)
            elif cmd in ('add_dependency', 'add_dep'):
                return line_magics.add_dep(line)
            elif cmd == 'turn_off_warnings_for':
                return line_magics.turn_off_warnings_for(line)
            elif cmd == 'turn_on_warnings_for':
                return line_magics.turn_on_warnings_for(line)
            elif cmd in ('mode', 'exec_mode'):
                return line_magics.set_exec_mode(line)
            elif cmd in line_magic_names:
                logger.warning('We have a magic for %s, but have not yet registered it', cmd)
                return None
            else:
                logger.warning(line_magics.USAGE)
                return None

        def _safety(line: str):
            # this is to avoid capturing `self` and creating an extra reference to the singleton
            try:
                cmd, line = line.split(' ', 1)
            except ValueError:
                cmd, line = line, ''
            try:
                line, fname = line.split('>', 1)
            except ValueError:
                line, fname = line, None
            line = line.strip()
            if fname is not None:
                fname = fname.strip()

            outstr = _handle(cmd, line)
            if outstr is None:
                return

            if fname is None:
                print_(outstr)
            else:
                with open(fname, 'w') as f:
                    f.write(outstr)

        # FIXME (smacke): probably not a great idea to rely on this
        _safety.__name__ = _SAFETY_LINE_MAGIC
        return register_line_magic(_safety)

    @property
    def dependency_tracking_enabled(self):
        return self.settings.track_dependencies

    @property
    def cell_magic_name(self):
        return self._cell_magic.__name__

    @property
    def line_magic_name(self):
        return self._line_magic.__name__

    def all_data_symbols(self) -> Iterable[DataSymbol]:
        for alias_set in self.aliases.values():
            yield from alias_set

    def test_and_clear_detected_flag(self):
        ret = self.safety_issue_detected
        self.safety_issue_detected = False
        return ret

    def _gc(self):
        # Need to do the garbage check and the collection separately
        garbage_syms = [dsym for dsym in self.all_data_symbols() if dsym.is_new_garbage()]
        for dsym in garbage_syms:
            dsym.collect_self_garbage()

    def retrieve_namespace_attr_or_sub(self, obj: Any, attr_or_sub: SupportedIndexType, is_subscript: bool):
        try:
            if is_subscript:
                # TODO: more complete list of things that are checkable
                #  or could cause side effects upon subscripting
                return obj[attr_or_sub]
            else:
                if self.is_develop:
                    assert isinstance(attr_or_sub, str)
                return getattr(obj, cast(str, attr_or_sub))
        except (AttributeError, IndexError, KeyError):
            raise
        except Exception as e:
            if self.is_develop:
                logger.warning('unexpected exception: %s', e)
                logger.warning('object: %s', obj)
                logger.warning('attr / subscript: %s', attr_or_sub)
            raise e



# -*- coding: future_annotations -*-
import ast
import logging
import sys
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Set, Tuple

logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)


class ContainsNamedExprVisitor(ast.NodeVisitor):
    def __init__(self):
        self.contains_named_expr = False

    def __call__(self, node: ast.stmt) -> bool:
        if sys.version_info.minor < 8:
            return False
        self.visit(node)
        return self.contains_named_expr

    def visit_NamedExpr(self, node):
        self.contains_named_expr = True

    def generic_visit(self, node: ast.AST):
        if self.contains_named_expr:
            return
        super().generic_visit(node)


def stmt_contains_lval(node: ast.stmt):
    # TODO: expand to method calls, etc.
    simple_contains_lval = isinstance(node, (
        ast.Assign,
        ast.AnnAssign,
        ast.AugAssign,
        ast.ClassDef,
        ast.FunctionDef,
        ast.AsyncFunctionDef,
        ast.For,
        ast.Import,
        ast.ImportFrom,
        ast.With,
    ))
    return simple_contains_lval or ContainsNamedExprVisitor()(node)



# -*- coding: future_annotations -*-
import ast
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Union


EMIT_EVENT = '_X5ix_NBSAFETY_EVT_EMIT'
TRACING_ENABLED = '_X5ix_NBSAFETY_TRACING_ENABLED'


def make_loop_iter_flag_name(loop_node: Union[int, ast.AST]):
    loop_node_id = loop_node if isinstance(loop_node, int) else id(loop_node)
    return '_X5ix_NBSAFETY_LOOPED_ONCE_{}'.format(loop_node_id)



# -*- coding: future_annotations -*-
import ast
from contextlib import contextmanager
import logging
from typing import cast, TYPE_CHECKING
import sys

from nbsafety.analysis.attr_symbols import resolve_slice_to_constant
from nbsafety.extra_builtins import EMIT_EVENT
from nbsafety.tracing.trace_events import TraceEvent
from nbsafety.utils import fast

if TYPE_CHECKING:
    from typing import Dict, List, Optional, Set, Union


logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)


class AstEavesdropper(ast.NodeTransformer):
    def __init__(self, orig_to_copy_mapping: Dict[int, ast.AST]):
        self._orig_to_copy_mapping = orig_to_copy_mapping
        self._top_level_node_for_symbol: Optional[ast.AST] = None

    def _emitter_ast(self):
        return fast.Name(EMIT_EVENT, ast.Load())

    def _get_copy_id_ast(self, orig_node_id: Union[int, ast.AST]):
        if not isinstance(orig_node_id, int):
            orig_node_id = id(orig_node_id)
        return fast.Num(id(self._orig_to_copy_mapping[orig_node_id]))

    def _make_tuple_event_for(self, node: ast.AST, event: TraceEvent, orig_node_id=None, **kwargs):
        with fast.location_of(node):
            tuple_node = fast.Tuple(
                [
                    fast.Call(
                        func=self._emitter_ast(),
                        args=[event.to_ast(), self._get_copy_id_ast(orig_node_id or node)],
                        keywords=[] if len(kwargs) == 0 else fast.kwargs(**kwargs),
                    ),
                    node,
                ],
                ast.Load()
            )
            slc: Union[ast.Constant, ast.Num, ast.Index] = fast.Num(1)
            if sys.version_info < (3, 9):
                slc = fast.Index(slc)
            return fast.Subscript(tuple_node, slc, ast.Load())

    def visit(self, node: ast.AST):
        ret = super().visit(node)
        if isinstance(node, ast.stmt):
            # we haven't inserted statements yet, and StatementInserter needs the previous ids to be identical
            assert ret is node
        return ret

    @contextmanager
    def attrsub_context(self, top_level_node: Optional[ast.AST]):
        old = self._top_level_node_for_symbol
        if old is None or top_level_node is None:
            # entering context when we are already inside chain is a no-op,
            # but we can specify a context of not being in chain if we are
            # inside one (in order to support arguments)
            self._top_level_node_for_symbol = top_level_node
        yield
        self._top_level_node_for_symbol = old

    @property
    def _inside_attrsub_load_chain(self):
        return self._top_level_node_for_symbol is not None

    def visit_Attribute(self, node: ast.Attribute, call_context=False):
        with fast.location_of(node.value):
            attr_node = cast(ast.Attribute, node)
            attr_or_sub = fast.Str(attr_node.attr)
        return self.visit_Attribute_or_Subscript(node, attr_or_sub, call_context=call_context)

    def _maybe_convert_ast_subscript(self, subscript: ast.AST) -> ast.AST:
        if isinstance(subscript, ast.Index):
            return self.visit(subscript.value)  # type: ignore
        elif isinstance(subscript, ast.Slice):
            lower = fast.NameConstant(None) if subscript.lower is None else self.visit(subscript.lower)
            upper = fast.NameConstant(None) if subscript.upper is None else self.visit(subscript.upper)
            return fast.Call(
                func=fast.Name('slice', ast.Load()),
                args=[lower, upper] + ([] if subscript.step is None else [self.visit(subscript.step)]),
                keywords=[],
            )
        else:
            return subscript

    def visit_Subscript(self, node: ast.Subscript, call_context=False):
        with self.attrsub_context(None):
            with fast.location_of(node.slice if hasattr(node.slice, 'lineno') else node.value):
                slc = self._maybe_convert_ast_subscript(node.slice)
                if isinstance(slc, (ast.ExtSlice, ast.Tuple)):
                    elts = slc.elts if isinstance(slc, ast.Tuple) else slc.dims  # type: ignore
                    elts = [self._maybe_convert_ast_subscript(elt) for elt in elts]  # type: ignore
                    slc = fast.Tuple(elts, ast.Load())
                slc = fast.Call(
                    func=self._emitter_ast(),
                    args=[
                        TraceEvent.subscript_slice.to_ast(),
                        self._get_copy_id_ast(node),
                    ],
                    keywords=fast.kwargs(ret=cast(ast.expr, slc)),
                )
                replacement_slice = fast.Call(
                    func=self._emitter_ast(),
                    args=[
                        TraceEvent._load_saved_slice.to_ast(),
                        self._get_copy_id_ast(node.slice),
                    ],
                    keywords=[],
                )
                if sys.version_info >= (3, 9):
                    node.slice = replacement_slice
                else:
                    node.slice = fast.Index(replacement_slice)
        return self.visit_Attribute_or_Subscript(node, slc, call_context=call_context)

    def _maybe_wrap_symbol_in_before_after_tracing(
        self, node, call_context=False, orig_node_id=None, begin_kwargs=None, end_kwargs=None
    ):
        if self._inside_attrsub_load_chain:
            return node
        orig_node = node
        orig_node_id = orig_node_id or id(orig_node)
        begin_kwargs = begin_kwargs or {}
        end_kwargs = end_kwargs or {}

        ctx = getattr(orig_node, 'ctx', ast.Load())
        is_load = isinstance(ctx, ast.Load)

        with fast.location_of(node):
            begin_kwargs['ret'] = self._get_copy_id_ast(orig_node_id)
            if is_load:
                end_ret = orig_node
            elif isinstance(orig_node, (ast.Attribute, ast.Subscript)):
                end_ret = orig_node.value
            else:
                raise TypeError('Unsupported node type for before / after symbol tracing: %s', type(orig_node))
            end_kwargs['ret'] = end_ret
            end_kwargs['ctx'] = fast.Str(ctx.__class__.__name__)
            end_kwargs['call_context'] = fast.NameConstant(call_context)
            node = fast.Call(
                func=self._emitter_ast(),
                args=[
                    TraceEvent.after_complex_symbol.to_ast(),
                    fast.Call(
                        # this will return the node id
                        func=self._emitter_ast(),
                        args=[TraceEvent.before_complex_symbol.to_ast(), self._get_copy_id_ast(orig_node_id)],
                        keywords=fast.kwargs(**begin_kwargs),
                    )
                ],
                keywords=fast.kwargs(**end_kwargs),
            )
            if not is_load:
                if isinstance(orig_node, ast.Attribute):
                    node = fast.Attribute(
                        value=node,
                        attr=orig_node.attr,
                    )
                elif isinstance(orig_node, ast.Subscript):
                    node = fast.Subscript(
                        value=node,
                        slice=orig_node.slice,
                    )
                else:
                    raise TypeError(
                        'Symbol tracing stores unsupported for node %s with type %s' %(orig_node, type(orig_node))
                    )
                node.ctx = ast.Store()
        # end location_of(node)
        return node

    def visit_Attribute_or_Subscript(
        self,
        node: Union[ast.Attribute, ast.Subscript],
        attr_or_sub: ast.expr,
        call_context: bool = False
    ):
        orig_node_id = id(node)
        with fast.location_of(node.value):
            extra_args: List[ast.keyword] = []
            if isinstance(node.value, ast.Name):
                extra_args = fast.kwargs(obj_name=fast.Str(node.value.id))

            subscript_name = None
            if isinstance(node, ast.Subscript):
                slice_val = resolve_slice_to_constant(node)
                if isinstance(slice_val, ast.Name):
                    # TODO: this should be more general than just simple ast.Name subscripts
                    subscript_name = slice_val.id

            with self.attrsub_context(node):
                node.value = fast.Call(
                    func=self._emitter_ast(),
                    args=[
                        TraceEvent.subscript.to_ast() if isinstance(node, ast.Subscript) else TraceEvent.attribute.to_ast(),
                        self._get_copy_id_ast(node.value)
                    ],
                    keywords=fast.kwargs(
                        ret=self.visit(node.value),
                        attr_or_subscript=attr_or_sub,
                        ctx=fast.Str(node.ctx.__class__.__name__),
                        call_context=fast.NameConstant(call_context),
                        top_level_node_id=self._get_copy_id_ast(self._top_level_node_for_symbol),
                        subscript_name=(
                            fast.NameConstant(subscript_name) if subscript_name is None else fast.Str(subscript_name)
                        ),
                    ) + extra_args
                )
        # end fast.location_of(node.value)

        return self._maybe_wrap_symbol_in_before_after_tracing(node, orig_node_id=orig_node_id)

    def _get_replacement_args(self, args, keywords: bool):
        replacement_args = []
        for arg in args:
            is_starred = isinstance(arg, ast.Starred)
            is_kwstarred = keywords and arg.arg is None
            if keywords or is_starred:
                maybe_kwarg = getattr(arg, 'value')
            else:
                maybe_kwarg = arg
            with fast.location_of(maybe_kwarg):
                with self.attrsub_context(None):
                    visited_maybe_kwarg = self.visit(maybe_kwarg)
                with self.attrsub_context(None):
                    new_arg_value = cast(ast.expr, fast.Call(
                        func=self._emitter_ast(),
                        args=[TraceEvent.argument.to_ast(), self._get_copy_id_ast(maybe_kwarg)],
                        keywords=fast.kwargs(
                            ret=visited_maybe_kwarg,
                            is_starred=fast.NameConstant(is_starred),
                            is_kwstarred=fast.NameConstant(is_kwstarred)
                        ),
                    ))
            if keywords or is_starred:
                setattr(arg, 'value', new_arg_value)
            else:
                arg = new_arg_value
            replacement_args.append(arg)
        return replacement_args

    def visit_Call(self, node: ast.Call):
        orig_node_id = id(node)

        with self.attrsub_context(node):
            if isinstance(node.func, ast.Attribute):
                node.func = self.visit_Attribute(node.func, call_context=True)
            elif isinstance(node.func, ast.Subscript):
                node.func = self.visit_Subscript(node.func, call_context=True)
            else:
                node.func = self.visit(node.func)

        # TODO: need a way to rewrite ast of subscript args,
        #  and to process these separately from outer rewrite

        node.args = self._get_replacement_args(node.args, False)
        node.keywords = self._get_replacement_args(node.keywords, True)

        # in order to ensure that the args are processed with appropriate active scope,
        # we need to make sure not to use the active namespace scope on args (in the case
        # of a function call on an ast.Attribute).
        #
        # We do so by emitting an "enter argument list", whose handler pushes the current active
        # scope while we process each argument. The "end argument list" event will then restore
        # the active scope.
        #
        # This effectively rewrites function calls as follows:
        # f(a, b, ..., c) -> trace(f, 'enter argument list')(a, b, ..., c)
        with fast.location_of(node.func):
            node.func = fast.Call(
                func=self._emitter_ast(),
                args=[TraceEvent.before_call.to_ast(), self._get_copy_id_ast(orig_node_id)],
                keywords=fast.kwargs(
                    ret=node.func,
                    call_node_id=self._get_copy_id_ast(orig_node_id),
                ),
            )

        # f(a, b, ..., c) -> trace(f(a, b, ..., c), 'exit argument list')
        with fast.location_of(node):
            node = fast.Call(
                func=self._emitter_ast(),
                args=[TraceEvent.after_call.to_ast(), self._get_copy_id_ast(orig_node_id)],
                keywords=fast.kwargs(
                    ret=node,
                    call_node_id=self._get_copy_id_ast(orig_node_id),
                ),
            )

        return self._maybe_wrap_symbol_in_before_after_tracing(node, call_context=True, orig_node_id=orig_node_id)

    def visit_Assign(self, node: ast.Assign):
        new_targets = []
        for target in node.targets:
            new_targets.append(self.visit(target))
        node.targets = new_targets
        orig_value_id = id(node.value)
        with fast.location_of(node.value):
            node.value = self._make_tuple_event_for(
                self.visit(node.value), TraceEvent.before_assign_rhs, orig_node_id=orig_value_id
            )
            node.value = fast.Call(
                func=self._emitter_ast(),
                args=[TraceEvent.after_assign_rhs.to_ast(), self._get_copy_id_ast(orig_value_id)],
                keywords=fast.kwargs(ret=node.value),
            )
        return node

    def visit_Lambda(self, node: ast.Lambda):
        assert isinstance(getattr(node, 'ctx', ast.Load()), ast.Load)
        visited = self.generic_visit(node)
        with fast.location_of(node):
            subscripted_node = self._make_tuple_event_for(
                visited, TraceEvent.before_lambda, orig_node_id=id(node)
            )
            return fast.Call(
                func=self._emitter_ast(),
                args=[TraceEvent.after_lambda.to_ast(), self._get_copy_id_ast(node)],
                keywords=fast.kwargs(ret=subscripted_node),
            )

    @staticmethod
    def _ast_container_to_literal_trace_evt(
        node: Union[ast.Dict, ast.List, ast.Set, ast.Tuple], before: bool
    ) -> TraceEvent:
        if isinstance(node, ast.Dict):
            return TraceEvent.before_dict_literal if before else TraceEvent.after_dict_literal
        elif isinstance(node, ast.List):
            return TraceEvent.before_list_literal if before else TraceEvent.after_list_literal
        elif isinstance(node, ast.Set):
            return TraceEvent.before_set_literal if before else TraceEvent.after_set_literal
        elif isinstance(node, ast.Tuple):
            return TraceEvent.before_tuple_literal if before else TraceEvent.after_tuple_literal
        else:
            raise TypeError('invalid ast node: %s', ast.dump(node))

    def visit_literal(self, node: Union[ast.Dict, ast.List, ast.Set, ast.Tuple], should_inner_visit=True):
        maybe_visited: ast.AST = node
        if should_inner_visit:
            maybe_visited = self.generic_visit(node)
        if not isinstance(getattr(node, 'ctx', ast.Load()), ast.Load):
            return maybe_visited
        with fast.location_of(node):
            subscripted_node = self._make_tuple_event_for(
                maybe_visited, self._ast_container_to_literal_trace_evt(node, before=True), orig_node_id=id(node)
            )
            return fast.Call(
                func=self._emitter_ast(),
                args=[
                    self._ast_container_to_literal_trace_evt(node, before=False).to_ast(),
                    self._get_copy_id_ast(node)
                ],
                keywords=fast.kwargs(ret=subscripted_node),
            )

    def visit_List(self, node: ast.List):
        return self.visit_List_or_Set_or_Tuple(node)

    def visit_Set(self, node: ast.Set):
        return self.visit_List_or_Set_or_Tuple(node)

    def visit_Tuple(self, node: ast.Tuple):
        return self.visit_List_or_Set_or_Tuple(node)

    @staticmethod
    def _ast_container_to_elt_trace_evt(node: Union[ast.List, ast.Set, ast.Tuple]) -> TraceEvent:
        if isinstance(node, ast.List):
            return TraceEvent.list_elt
        elif isinstance(node, ast.Set):
            return TraceEvent.set_elt
        elif isinstance(node, ast.Tuple):
            return TraceEvent.tuple_elt
        else:
            raise TypeError('invalid ast node: %s', ast.dump(node))

    def visit_List_or_Set_or_Tuple(self, node: Union[ast.List, ast.Set, ast.Tuple]):
        traced_elts: List[ast.expr] = []
        is_load = isinstance(getattr(node, 'ctx', ast.Load()), ast.Load)
        saw_starred = False
        for i, elt in enumerate(node.elts):
            if isinstance(elt, ast.Starred):
                # TODO: trace starred elts too
                saw_starred = True
                traced_elts.append(elt)
                continue
            elif not is_load:
                traced_elts.append(self.visit(elt))
                continue
            with fast.location_of(elt):
                traced_elts.append(fast.Call(
                    func=self._emitter_ast(),
                    args=[
                        self._ast_container_to_elt_trace_evt(node).to_ast(),
                        self._get_copy_id_ast(elt),
                    ],
                    keywords=fast.kwargs(
                        ret=self.visit(elt),
                        index=fast.NameConstant(None) if saw_starred else fast.Num(i),
                        container_node_id=self._get_copy_id_ast(node),
                    )
                ))
        node.elts = traced_elts
        return self.visit_literal(node, should_inner_visit=False)

    def visit_Dict(self, node: ast.Dict):
        traced_keys: List[Optional[ast.expr]] = []
        traced_values: List[ast.expr] = []
        for k, v in zip(node.keys, node.values):
            is_dict_unpack = (k is None)
            if is_dict_unpack:
                traced_keys.append(None)
            else:
                with fast.location_of(k):
                    traced_keys.append(fast.Call(
                        func=self._emitter_ast(),
                        args=[TraceEvent.dict_key.to_ast(), self._get_copy_id_ast(k)],
                        keywords=fast.kwargs(
                            ret=self.visit(k),
                            value_node_id=self._get_copy_id_ast(v),
                            dict_node_id=self._get_copy_id_ast(node),
                        )
                    ))
            with fast.location_of(v):
                if is_dict_unpack:
                    key_node_id_ast = fast.NameConstant(None)
                else:
                    key_node_id_ast = self._get_copy_id_ast(k)
                traced_values.append(fast.Call(
                    func=self._emitter_ast(),
                    args=[TraceEvent.dict_value.to_ast(), self._get_copy_id_ast(v)],
                    keywords=fast.kwargs(
                        ret=self.visit(v),
                        key_node_id=key_node_id_ast,
                        dict_node_id=self._get_copy_id_ast(node),
                    )
                ))
        node.keys = traced_keys
        node.values = traced_values
        return self.visit_literal(node, should_inner_visit=False)

    def visit_Return(self, node: ast.Return):
        if node.value is None:
            return node
        with fast.location_of(node):
            node.value = fast.Call(
                func=self._emitter_ast(),
                args=[TraceEvent.after_return.to_ast(), self._get_copy_id_ast(node.value)],
                keywords=fast.kwargs(
                    ret=self._make_tuple_event_for(
                        self.visit(node.value),
                        TraceEvent.before_return,
                        orig_node_id=id(node.value),
                    ),
                ),
            )
        return node

    def visit_Delete(self, node: ast.Delete):
        ret = cast(ast.Delete, self.generic_visit(node))
        for target in ret.targets:
            target.ctx = ast.Del()  # type: ignore
        return ret



# -*- coding: future_annotations -*-
import ast
from collections import defaultdict
from enum import Enum
import logging
import sys
from typing import cast, TYPE_CHECKING

try:
    import numpy
except ImportError:
    numpy = None

from nbsafety.data_model import sizing
from nbsafety.data_model.annotation_utils import get_type_annotation, make_annotation_string
from nbsafety.data_model.code_cell import ExecutedCodeCell, cells
from nbsafety.data_model.timestamp import Timestamp
from nbsafety.data_model.update_protocol import UpdateProtocol
from nbsafety.singletons import nbs, tracer

if TYPE_CHECKING:
    from nbsafety.types import SupportedIndexType
    from typing import Any, Dict, List, Optional, Set

    # avoid circular imports
    from nbsafety.data_model.scope import Scope
    from nbsafety.data_model.namespace import Namespace

logger = logging.getLogger(__name__)
logger.setLevel(logging.ERROR)


class DataSymbolType(Enum):
    DEFAULT = 'default'
    SUBSCRIPT = 'subscript'
    FUNCTION = 'function'
    CLASS = 'class'
    IMPORT = 'import'
    ANONYMOUS = 'anonymous'


class DataSymbol:

    NULL = object()

    IMMUTABLE_TYPES = {
        bytes,
        bytearray,
        float,
        frozenset,
        int,
        str,
        tuple,
    }

    def __init__(
        self,
        name: SupportedIndexType,
        symbol_type: DataSymbolType,
        obj: Any,
        containing_scope: Scope,
        stmt_node: Optional[ast.AST] = None,
        # TODO: also keep a reference to the target node?
        refresh_cached_obj: bool = False,
        implicit: bool = False,
    ):
        if refresh_cached_obj:
            # TODO: clean up redundancies
            assert implicit
            assert stmt_node is None
        self.name = name
        self.symbol_type = symbol_type
        self.obj = obj
        self._tombstone = False
        self._cached_out_of_sync = True
        self.cached_obj_id = None
        self.cached_obj_type = None
        if refresh_cached_obj:
            self._refresh_cached_obj()
        self.containing_scope = containing_scope
        self.stmt_node = self.update_stmt_node(stmt_node)
        self._funcall_live_symbols = None
        self.parents: Dict[DataSymbol, List[Timestamp]] = defaultdict(list)
        self.children: Dict[DataSymbol, List[Timestamp]] = defaultdict(list)

        self.call_scope: Optional[Scope] = None
        if self.is_function:
            self.call_scope = self.containing_scope.make_child_scope(self.name)

        # initialize at -1 since the corresponding piece of data could already be around,
        # and we don't want liveness checker to think this was newly created unless we
        # explicitly trace an update somewhere
        self._timestamp: Timestamp = Timestamp.uninitialized()
        # The version is a simple counter not associated with cells that is bumped whenever the timestamp is updated
        self._version: int = 0
        self._defined_cell_num = cells().exec_counter()

        # The necessary last-updated timestamp / cell counter for this symbol to not be stale
        self.required_timestamp: Timestamp = self.timestamp

        # for each usage of this dsym, the version that was used, if different from the timestamp of usage
        self.timestamp_by_used_time: Dict[Timestamp, Timestamp] = {}
        # History of definitions at time of liveness
        self.timestamp_by_liveness_time: Dict[Timestamp, Timestamp] = {}
        # All timestamps associated with this symbol
        self.updated_timestamps: Set[Timestamp] = set()

        self.fresher_ancestors: Set[DataSymbol] = set()
        self.fresher_ancestor_timestamps: Set[Timestamp] = set()

        # cells where this symbol was live
        self.cells_where_deep_live: Set[ExecutedCodeCell] = set()
        self.cells_where_shallow_live: Set[ExecutedCodeCell] = set()

        self._last_computed_staleness_cache_ts: int = -1
        self._is_stale_at_position_cache: Dict[int, bool] = {}

        # if implicitly created when tracing non-store-context ast nodes
        self._implicit = implicit

        # Will never be stale if no_warning is True
        self.disable_warnings = False
        self._temp_disable_warnings = False

        nbs().aliases[id(obj)].add(self)
        if isinstance(self.name, str) and not self.is_anonymous and not self.containing_scope.is_namespace_scope:
            ns = self.namespace
            if ns is not None and ns.scope_name == 'self':
                # hack to get a better name than `self.whatever` for fields of this object
                # not ideal because it relies on the `self` convention but is probably
                # acceptable for the use case of improving readable names
                ns.scope_name = self.name

    @property
    def cells_where_live(self) -> Set[ExecutedCodeCell]:
        return self.cells_where_deep_live | self.cells_where_shallow_live

    def __repr__(self) -> str:
        return f'<{self.readable_name}>'

    def __str__(self) -> str:
        return self.readable_name

    def __hash__(self):
        return hash(self.full_path)

    def temporary_disable_warnings(self):
        self._temp_disable_warnings = True

    @property
    def last_used_timestamp(self):
        if len(self.timestamp_by_used_time) == 0:
            return Timestamp.uninitialized()
        else:
            return max(self.timestamp_by_used_time.keys())

    @property
    def namespace_stale_symbols(self) -> Set[DataSymbol]:
        ns = self.namespace
        return set() if ns is None else ns.namespace_stale_symbols

    @property
    def timestamp_excluding_ns_descendents(self):
        return self._timestamp

    @property
    def timestamp(self) -> Timestamp:
        ts = self._timestamp
        ns = self.namespace
        return ts if ns is None else max(ts, ns.max_descendent_timestamp)

    @property
    def defined_cell_num(self) -> int:
        return self._defined_cell_num

    @property
    def readable_name(self) -> str:
        return self.containing_scope.make_namespace_qualified_name(self)

    @property
    def is_subscript(self):
        return self.symbol_type == DataSymbolType.SUBSCRIPT

    @property
    def is_class(self):
        return self.symbol_type == DataSymbolType.CLASS

    @property
    def is_function(self):
        return self.symbol_type == DataSymbolType.FUNCTION

    @property
    def is_import(self):
        return self.symbol_type == DataSymbolType.IMPORT

    @property
    def imported_module(self) -> str:
        if not self.is_import:
            raise ValueError('only IMPORT symbols have `imported_module` property')
        if isinstance(self.stmt_node, ast.Import):
            for alias in self.stmt_node.names:
                name = alias.asname or alias.name
                if name == self.name:
                    return alias.name
            raise ValueError('Unable to find module for symbol %s is stmt %s' % (self, ast.dump(self.stmt_node)))
        elif isinstance(self.stmt_node, ast.ImportFrom):
            return self.stmt_node.module
        else:
            raise TypeError('Invalid stmt type for import symbol: %s' % ast.dump(self.stmt_node))

    @property
    def imported_symbol_original_name(self) -> str:
        if not self.is_import:
            raise ValueError('only IMPORT symbols have `imported_symbol_original_name` property')
        if isinstance(self.stmt_node, ast.Import):
            return self.imported_module
        elif isinstance(self.stmt_node, ast.ImportFrom):
            for alias in self.stmt_node.names:
                name = alias.asname or alias.name
                if name == self.name:
                    return alias.name
            raise ValueError('Unable to find module for symbol %s is stmt %s' % (self, ast.dump(self.stmt_node)))
        else:
            raise TypeError('Invalid stmt type for import symbol: %s' % ast.dump(self.stmt_node))

    def get_top_level(self) -> Optional[DataSymbol]:
        if not self.containing_scope.is_namespace_scope:
            return self
        else:
            containing_scope = cast('Namespace', self.containing_scope)
            for alias in nbs().aliases[containing_scope.obj_id]:
                if alias.is_globally_accessible:
                    return alias.get_top_level()
            return None

    def get_import_string(self) -> str:
        if not self.is_import:
            raise ValueError('only IMPORT symbols support recreating the import string')
        module = self.imported_module
        if isinstance(self.stmt_node, ast.Import):
            if module == self.name:
                return f'import {module}'
            else:
                return f'import {module} as {self.name}'
        elif isinstance(self.stmt_node, ast.ImportFrom):
            original_symbol_name = self.imported_symbol_original_name
            if original_symbol_name == self.name:
                return f'from {module} import {original_symbol_name}'
            else:
                return f'from {module} import {original_symbol_name} as {self.name}'
        else:
            raise TypeError('Invalid stmt type for import symbol: %s' % ast.dump(self.stmt_node))

    @property
    def is_anonymous(self):
        return self.symbol_type == DataSymbolType.ANONYMOUS

    @property
    def is_implicit(self):
        return self._implicit

    def shallow_clone(self, new_obj, new_containing_scope, symbol_type):
        return self.__class__(self.name, symbol_type, new_obj, new_containing_scope)

    @property
    def obj_id(self):
        return id(self.obj)

    @property
    def obj_type(self):
        return type(self.obj)

    def get_type_annotation(self):
        return get_type_annotation(self.obj)

    def get_type_annotation_string(self) -> str:
        return make_annotation_string(self.get_type_annotation())

    @property
    def namespace(self):
        return nbs().namespaces.get(self.obj_id, None)

    @property
    def containing_namespace(self) -> Optional[Namespace]:
        if self.containing_scope.is_namespace_scope:
            return cast('Namespace', self.containing_scope)
        else:
            return None

    @property
    def full_path(self):
        return self.containing_scope.full_path + (self.name,)

    @property
    def full_namespace_path(self):
        return self.containing_scope.make_namespace_qualified_name(self)

    @property
    def is_garbage(self):
        return self._tombstone

    def is_new_garbage(self):
        if self._tombstone:
            return False
        if numpy is not None and self.containing_namespace is not None and isinstance(self.containing_namespace.obj, numpy.ndarray):
            # numpy atoms are not interned (so assigning array elts to a variable does not bump refcount);
            # also seems that refcount is always 0, so just check if the containing namespace is garbage
            return self.containing_namespace.is_garbage
        return self.get_ref_count() == 0

    @property
    def is_globally_accessible(self):
        return self.containing_scope.is_globally_accessible

    @property
    def is_user_accessible(self):
        return (
            self.is_globally_accessible
            and not self.is_anonymous
            and not (
                self.containing_namespace is not None
                and self.containing_namespace.is_anonymous
            )
        )

    def collect_self_garbage(self):
        """
        Just null out the reference to obj; we need to keep the edges
        and namespace relationships around for staleness propagation.
        """
        # TODO: ideally we should figure out how to GC the symbols themselves
        #  and remove them from the symbol graph, to keep this from getting
        #  too large. One idea is that a symbol can be GC'd if its reachable
        #  descendants are all tombstone'd, and likewise a namespace can be
        #  GC'd if all of its children are GC'able as per prior criterion.
        self._tombstone = True
        ns = nbs().namespaces.get(self.obj_id, None)
        if ns is not None:
            ns._tombstone = True
            ns.obj = None
        self.obj = None

    # def update_type(self, new_type):
    #     self.symbol_type = new_type
    #     if self.is_function:
    #         self.call_scope = self.containing_scope.make_child_scope(self.name)
    #     else:
    #         self.call_scope = None

    def update_obj_ref(self, obj, refresh_cached=True):
        logger.info("%s update obj ref to %s", self, obj)
        self._tombstone = False
        self._cached_out_of_sync = True
        if nbs().settings.mark_typecheck_failures_unsafe and self.cached_obj_type != type(obj):
            for cell in self.cells_where_live:
                cell.invalidate_typecheck_result()
        self.cells_where_shallow_live.clear()
        self.cells_where_deep_live.clear()
        self.obj = obj
        if self.cached_obj_id is not None and self.cached_obj_id != self.obj_id:
            new_ns = nbs().namespaces.get(self.obj_id, None)
            # don't overwrite existing namespace for this obj
            old_ns = nbs().namespaces.get(self.cached_obj_id, None)
            if old_ns is not None and (new_ns is None or not new_ns.max_descendent_timestamp.is_initialized) and old_ns.full_namespace_path == self.full_namespace_path:
                if new_ns is None:
                    logger.info("create fresh copy of namespace %s", old_ns)
                    new_ns = old_ns.fresh_copy(obj)
                else:
                    new_ns.scope_name = old_ns.scope_name
                    new_ns.parent_scope = old_ns.parent_scope
                old_ns.transfer_symbols_to(new_ns)
            self._handle_aliases()
        if refresh_cached:
            self._refresh_cached_obj()

    def invalidate_cached(self):
        self._cached_out_of_sync = True
        self.cached_obj_id = None
        self.cached_obj_type = None

    def get_ref_count(self):
        if self.obj is None or self.obj is DataSymbol.NULL:
            return -1
        total = sys.getrefcount(self.obj) - 1
        total -= len(nbs().aliases[self.obj_id])
        ns = nbs().namespaces.get(self.obj_id, None)
        if ns is not None and ns.obj is not None and ns.obj is not DataSymbol.NULL:
            total -= 1
        return total

    def prev_obj_definitely_equal_to_current_obj(self, prev_obj: Optional[Any]) -> bool:
        if prev_obj is None:
            return False
        if not self._cached_out_of_sync or self.obj_id == self.cached_obj_id:
            return True
        if self.obj is None or prev_obj is DataSymbol.NULL:
            return self.obj is None and prev_obj is DataSymbol.NULL
        obj_type = type(self.obj)
        prev_type = type(prev_obj)
        if obj_type != prev_type:
            return False
        obj_size_ubound = sizing.sizeof(self.obj)
        if obj_size_ubound > sizing.MAX_SIZE:
            return False
        cached_obj_size_ubound = sizing.sizeof(prev_obj)
        if cached_obj_size_ubound > sizing.MAX_SIZE:
            return False
        return (obj_size_ubound == cached_obj_size_ubound) and self.obj == prev_obj

    def _handle_aliases(self):
        old_aliases = nbs().aliases.get(self.cached_obj_id, None)
        if old_aliases is not None:
            old_aliases.discard(self)
            if len(old_aliases) == 0:
                del nbs().aliases[self.cached_obj_id]
        nbs().aliases[self.obj_id].add(self)

    def update_stmt_node(self, stmt_node):
        self.stmt_node = stmt_node
        self._funcall_live_symbols = None
        if self.is_function:
            # TODO: in the case of lambdas, there will not necessarily be one
            #  symbol for a given statement. We need a more precise way to determine
            #  the symbol being called than by looking at the stmt in question.
            nbs().statement_to_func_cell[id(stmt_node)] = self
            self.call_scope = self.containing_scope.make_child_scope(self.name)
        return stmt_node

    def _refresh_cached_obj(self):
        self._cached_out_of_sync = False
        # don't keep an actual ref to avoid bumping prefcount
        self.cached_obj_id = self.obj_id
        self.cached_obj_type = self.obj_type

    def get_definition_args(self) -> List[str]:
        assert self.is_function and isinstance(self.stmt_node, (ast.FunctionDef, ast.AsyncFunctionDef, ast.Lambda))
        args = []
        for arg in self.stmt_node.args.args + self.stmt_node.args.kwonlyargs:
            args.append(arg.arg)
        if self.stmt_node.args.vararg is not None:
            args.append(self.stmt_node.args.vararg.arg)
        if self.stmt_node.args.kwarg is not None:
            args.append(self.stmt_node.args.kwarg.arg)
        return args

    def _match_call_args_with_definition_args(self):
        assert self.is_function and isinstance(self.stmt_node, (ast.FunctionDef, ast.AsyncFunctionDef, ast.Lambda))
        caller_node = self._get_calling_ast_node()
        if caller_node is None or not isinstance(caller_node, ast.Call):
            return []
        def_args = self.stmt_node.args.args
        if len(self.stmt_node.args.defaults) > 0:
            def_args = def_args[:-len(self.stmt_node.args.defaults)]
        if len(def_args) > 0 and def_args[0].arg == 'self':
            # FIXME: this is bad and I should feel bad
            def_args = def_args[1:]
        for def_arg, call_arg in zip(def_args, caller_node.args):
            if isinstance(call_arg, ast.Starred):
                # give up
                # TODO: handle this case
                break
            yield def_arg.arg, tracer().resolve_loaded_symbols(call_arg)
        seen_keys = set()
        for keyword in caller_node.keywords:
            key, value = keyword.arg, keyword.value
            if value is None:
                continue
            seen_keys.add(key)
            yield key, tracer().resolve_loaded_symbols(value)
        for key, value in zip(self.stmt_node.args.args[-len(self.stmt_node.args.defaults):], self.stmt_node.args.defaults):
            if key.arg in seen_keys:
                continue
            yield key.arg, tracer().resolve_loaded_symbols(value)

    def _get_calling_ast_node(self) -> Optional[ast.AST]:
        if isinstance(self.stmt_node, (ast.FunctionDef, ast.AsyncFunctionDef)):
            if self.name in ('__getitem__', '__setitem__', '__delitem__'):
                # TODO: handle case where we're looking for a subscript for the calling node
                return None
            for decorator in self.stmt_node.decorator_list:
                if isinstance(decorator, ast.Name) and decorator.id == 'property':
                    # TODO: handle case where we're looking for an attribute for the calling node
                    return None
        lexical_call_stack = tracer().lexical_call_stack
        if len(lexical_call_stack) == 0:
            return None
        prev_node_id_in_cur_frame_lexical = lexical_call_stack.get_field('prev_node_id_in_cur_frame_lexical')
        caller_ast_node = nbs().ast_node_by_id.get(prev_node_id_in_cur_frame_lexical, None)
        if caller_ast_node is None or not isinstance(caller_ast_node, ast.Call):
            return None
        return caller_ast_node

    def create_symbols_for_call_args(self) -> None:
        assert self.is_function
        seen_def_args = set()
        logger.info('create symbols for call to %s', self)
        for def_arg, deps in self._match_call_args_with_definition_args():
            seen_def_args.add(def_arg)
            # TODO: ideally we should try pass the actual objects to the DataSymbol ctor.
            #  Will require matching the signature with the actual call,
            #  which will be tricky I guess.
            self.call_scope.upsert_data_symbol_for_name(def_arg, None, deps, self.stmt_node, propagate=False)
            logger.info('def arg %s matched with deps %s', def_arg, deps)
        for def_arg in self.get_definition_args():
            if def_arg in seen_def_args:
                continue
            self.call_scope.upsert_data_symbol_for_name(def_arg, None, set(), self.stmt_node, propagate=False)

    @property
    def is_stale(self):
        if self.disable_warnings or self._temp_disable_warnings:
            return False
        return self.timestamp < self.required_timestamp or len(self.namespace_stale_symbols) > 0

    def _is_stale_at_position_impl(self, pos: int) -> bool:
        for par, timestamps in self.parents.items():
            for ts in timestamps:
                dep_introduced_pos = cells().from_timestamp(ts).position
                if dep_introduced_pos > pos:
                    continue
                for updated_ts in par.updated_timestamps:
                    if cells().from_timestamp(updated_ts).position > dep_introduced_pos:
                        continue
                    if updated_ts.cell_num > ts.cell_num or par.is_stale_at_position(ts.cell_num):
                        # logger.error("sym: %s", self)
                        # logger.error("pos: %s", pos)
                        # logger.error("parent: %s", par)
                        # logger.error("dep introdced ts: %s", ts)
                        # logger.error("dep introdced pos: %s", dep_introduced_pos)
                        # logger.error("par updated ts: %s", updated_ts)
                        # logger.error("par updated position: %s", cells().from_timestamp(updated_ts).position)
                        return True
        for sym in self.namespace_stale_symbols:
            if sym.is_stale_at_position(ts.cell_num):
                return True
        return False

    def is_stale_at_position(self, pos: int) -> bool:
        if not self.is_stale:
            return False
        if cells().exec_counter() > self._last_computed_staleness_cache_ts:
            self._is_stale_at_position_cache.clear()
            self._last_computed_staleness_cache_ts = cells().exec_counter()
        if pos in self._is_stale_at_position_cache:
            return self._is_stale_at_position_cache[pos]
        is_stale = self._is_stale_at_position_impl(pos)
        self._is_stale_at_position_cache[pos] = is_stale
        return is_stale



    def should_mark_stale(self, updated_dep):
        if self.disable_warnings:
            return False
        if updated_dep is self:
            return False
        return True

    def _is_simple_assign(self, new_deps: Set[DataSymbol]) -> bool:
        if not isinstance(self.stmt_node, (ast.Assign, ast.AnnAssign)):
            return False
        if len(new_deps) != 1:
            return False
        only_dep: DataSymbol = next(iter(new_deps))
        # obj ids can get reused for anon symbols like literals
        return not only_dep.is_anonymous and self.cached_obj_id == only_dep.obj_id

    def update_deps(
        self,
        new_deps: Set[DataSymbol],
        prev_obj: Any = None,
        overwrite: bool = True,
        mutated: bool = False,
        deleted: bool = False,
        propagate_to_namespace_descendents: bool = False,
        propagate: bool = True,
        refresh: bool = True,
    ) -> None:
        if self.is_import and self.obj_id == self.cached_obj_id:
            # skip updates for imported symbols
            # just bump the version if it's newly created
            if mutated or not self._timestamp.is_initialized:
                self._timestamp = Timestamp.current()
            return
        if mutated and self.obj_type in self.IMMUTABLE_TYPES:
            return
        # if we get here, no longer implicit
        self._implicit = False
        # quick last fix to avoid overwriting if we appear inside the set of deps to add (or a 1st order ancestor)
        # TODO: check higher-order ancestors too?
        overwrite = overwrite and self not in new_deps
        overwrite = overwrite and not any(self in new_dep.parents for new_dep in new_deps)
        logger.warning("symbol %s new deps %s", self, new_deps)
        new_deps.discard(self)
        if overwrite:
            for parent in self.parents.keys() - new_deps:
                parent.children.pop(self, None)
            self.parents.clear()

        for new_parent in new_deps - self.parents.keys():
            if new_parent is None:
                continue
            new_parent.children[self].append(Timestamp.current())
            self.parents[new_parent].append(Timestamp.current())
        self.required_timestamp = Timestamp.uninitialized()
        self.fresher_ancestors.clear()
        self.fresher_ancestor_timestamps.clear()
        if mutated or isinstance(self.stmt_node, ast.AugAssign):
            self.update_usage_info()
        equal_to_old = not mutated and self.prev_obj_definitely_equal_to_current_obj(prev_obj)
        if refresh:
            self.refresh(
                bump_version=not equal_to_old or type(self.obj) not in DataSymbol.IMMUTABLE_TYPES,
                # rationale: if this is a mutation for which we have more precise information,
                # then we don't need to update the ns descendents as this will already have happened.
                # also don't update ns descendents for things like `a = b`
                refresh_descendent_namespaces=not (mutated and not propagate_to_namespace_descendents) and not self._is_simple_assign(new_deps),
                refresh_namespace_stale=not mutated,
            )
        if propagate and (deleted or not equal_to_old):
            UpdateProtocol(self)(new_deps, mutated, propagate_to_namespace_descendents, refresh)
        self._refresh_cached_obj()
        nbs().updated_symbols.add(self)
        if self.is_class:
            # pop pending class defs and update obj ref
            pending_class_ns = tracer().pending_class_namespaces.pop()
            pending_class_ns.update_obj_ref(self.obj)

    def update_usage_info(self, used_time: Optional[Timestamp] = None, exclude_ns: bool = False) -> None:
        if used_time is None:
            used_time = Timestamp.current()
        if nbs().is_develop:
            logger.info('sym `%s` used in cell %d last updated in cell %d', self, used_time.cell_num, self.timestamp)
        if used_time not in self.timestamp_by_used_time and self.timestamp < used_time:
            self.timestamp_by_used_time[used_time] = (
                self.timestamp_excluding_ns_descendents if exclude_ns else self.timestamp
            )

    def refresh(
        self,
        bump_version=True,
        refresh_descendent_namespaces=False,
        refresh_namespace_stale=True,
        timestamp: Optional[Timestamp] = None,
        seen: Set[DataSymbol] = None,
    ) -> None:
        self._temp_disable_warnings = False
        if bump_version:
            self._timestamp = Timestamp.current() if timestamp is None else timestamp
            for cell in self.cells_where_live:
                cell.add_used_cell_counter(self, self._timestamp.cell_num)
            ns = self.containing_namespace
            if ns is not None:
                # logger.error("bump version of %s due to %s (value %s)", ns.full_path, self.full_path, self.obj)
                ns.max_descendent_timestamp = self._timestamp
                for alias in nbs().aliases[ns.obj_id]:
                    for cell in alias.cells_where_deep_live:
                        cell.add_used_cell_counter(alias, self._timestamp.cell_num)
            self.updated_timestamps.add(self._timestamp)
            self._version += 1
        if refresh_descendent_namespaces:
            if seen is None:
                seen = set()
            if self in seen:
                return
            seen.add(self)
            ns = self.namespace
            if ns is not None:
                for dsym in ns.all_data_symbols_this_indentation(exclude_class=True):
                    # this is to handle cases like `x = x.mutate(42)`, where
                    # we could have changed some member of x but returned the
                    # original object -- in this case, just assume that all
                    # the stale namespace descendents are no longer stale, as
                    # this is likely the user intention. For an example, see
                    # `test_external_object_update_propagates_to_stale_namespace_symbols()`
                    # in `test_multicell_precheck.py`
                    if not dsym.is_stale or refresh_namespace_stale:
                        # logger.error("refresh %s due to %s (value %s) via namespace %s", dsym.full_path, self.full_path, self.obj, ns.full_path)
                        dsym.refresh(refresh_descendent_namespaces=True, timestamp=self._timestamp, seen=seen)
            if refresh_namespace_stale:
                self.namespace_stale_symbols.clear()



# -*- coding: future_annotations -*-
from collections import defaultdict
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Any, List, Mapping, Optional, Sequence


class KeyDict(dict):
    def __missing__(self, key):
        return key


class GetterFallback:
    def __init__(self, stages: Sequence[Mapping[Any, Any]]):
        self.stages = stages

    def __getitem__(self, item):
        for stage in self.stages:
            if item in stage:
                return stage[item]
        raise KeyError()

    def __setitem__(self, key, value):
        return NotImplemented

    def __contains__(self, item):
        for stage in self.stages:
            if item in stage:
                return True
        return False



# -*- coding: future_annotations -*-
from enum import Enum
import os


class SafetyRunMode(Enum):
    PRODUCTION = 'NBSAFETY_PRODUCTION'
    DEVELOP = 'NBSAFETY_DEVELOP'

    @staticmethod
    def get():
        if SafetyRunMode.DEVELOP.value in os.environ and str(os.environ[SafetyRunMode.DEVELOP.value]) == '1':
            return SafetyRunMode.DEVELOP
        else:
            return SafetyRunMode.PRODUCTION


class ExecutionMode(Enum):
    NORMAL = 'normal'
    REACTIVE = 'reactive'



# -*- coding: future_annotations -*-
import functools
import logging

from nbsafety.data_model.code_cell import cells
from nbsafety.singletons import nbs
from test.utils import make_safety_fixture

logging.basicConfig(level=logging.ERROR)

# Reset dependency graph before each test
# _safety_fixture, run_cell_ = make_safety_fixture(trace_messages_enabled=True)
_safety_fixture, run_cell_ = make_safety_fixture()


def dynamic_or_static_only_test(test_f, dynamic_enabled, static_enabled):
    @functools.wraps(test_f)
    def dynamic_test_f(*args, **kwargs):
        orig_dynamic_enabled = nbs().mut_settings.dynamic_slicing_enabled
        orig_static_enabled = nbs().mut_settings.dynamic_slicing_enabled
        try:
            nbs().mut_settings.dynamic_slicing_enabled = dynamic_enabled
            nbs().mut_settings.static_slicing_enabled = static_enabled
            test_f(*args, **kwargs)
        finally:
            nbs().mut_settings.dynamic_slicing_enabled = orig_dynamic_enabled
            nbs().mut_settings.static_slicing_enabled = orig_static_enabled
    return dynamic_test_f


def dynamic_only_test(test_f):
    return dynamic_or_static_only_test(test_f, True, False)


def static_only_test(test_f):
    return dynamic_or_static_only_test(test_f, False, True)


def run_cell(cell):
    # print()
    # print('*******************************************')
    # print('running', cell)
    # print('*******************************************')
    # print()
    run_cell_(cell)


def compute_slice_stmts(cell_num):
    return cells().from_timestamp(cell_num).compute_slice_stmts()


def compute_slice(cell_num):
    return cells().from_timestamp(cell_num).compute_slice()


def num_stmts_in_slice(cell_num: int) -> int:
    return sum(len(stmts) for stmts in compute_slice_stmts(cell_num).values())


def test_simple():
    run_cell('a = 1')
    run_cell('b = 2')
    run_cell('logging.info(a)')
    run_cell('c = a + b')
    deps = set(compute_slice(4).keys())
    assert deps == {1, 2, 4}, 'got %s' % deps


def test_dynamic_symbol_usage():
    run_cell('x = 5')
    run_cell("""
class Foo:
    def foo(self):
        return x
""")
    run_cell("""
def foo():
    return Foo()
""")
    run_cell('logging.info(foo().foo())')
    deps = set(compute_slice(4).keys())
    assert deps == {1, 2, 3, 4}, 'got %s' % deps
    slice_size = num_stmts_in_slice(4)
    assert slice_size == 4, 'got %d' % slice_size


def test_nested_symbol_usage():
    run_cell('lst = [1, 2, 3, 4, 5]')
    run_cell('lst[1] = 3')
    run_cell('logging.info(lst[1])')
    deps = set(compute_slice(3).keys())
    assert deps == {1, 2, 3}, 'got %s' % deps
    slice_size = num_stmts_in_slice(3)
    assert slice_size == 3, 'got %d' % slice_size


def test_nested_symbol_usage_with_variable_subscript():
    for static_slicing_enabled in [True, False]:
        if not static_slicing_enabled and not nbs().mut_settings.dynamic_slicing_enabled:
            continue
        orig_enabled = nbs().mut_settings.static_slicing_enabled
        try:
            nbs().mut_settings.static_slicing_enabled = static_slicing_enabled
            run_cell('x = 1')
            run_cell('lst = [1, 2, 3, 4, 5]')
            run_cell('lst[x] = 3')
            run_cell('logging.info(lst[1])')
            deps = set(compute_slice(4).keys())
            assert deps == {1, 2, 3, 4}, 'got %s' % deps
            slice_size = num_stmts_in_slice(4)
            assert slice_size == 4, 'got %d' % slice_size
        finally:
            nbs().mut_settings.static_slicing_enabled = orig_enabled


@static_only_test
def test_liveness_timestamps():
    run_cell("""
x = 0
if True:
    y = 1
else:
    y = 2
""")
    run_cell('z = 42')
    run_cell('logging.info(x + 1)')
    deps = set(compute_slice(3).keys())
    assert deps == {1, 3}, 'got %s' % deps
    slice_size = num_stmts_in_slice(3)
    assert slice_size == 2, 'got %d' % slice_size


def test_list_mutations():
    run_cell('dummy = 0')
    run_cell('lst = []')
    run_cell('lst.append(1)')
    run_cell('lst.append(2)')
    run_cell('lst.append(3); lst.append(4)')
    run_cell('logging.info(lst)')
    deps = set(compute_slice(5).keys())
    assert deps == {2, 3, 4, 5}, 'got %s' % deps
    slice_size = num_stmts_in_slice(5)
    assert slice_size == 5, 'got %d' % slice_size


def test_imports():
    run_cell('import numpy as np')
    run_cell('arr = np.zeros((5,))')
    run_cell('logging.info(arr * 3)')
    deps = set(compute_slice(3).keys())
    assert deps == {1, 2, 3}, 'got %s' % deps
    slice_size = num_stmts_in_slice(3)
    assert slice_size == 3, 'got %d' % slice_size


def test_handle_stale():
    run_cell('a = 1')
    run_cell('b = 2 * a')
    run_cell('a = 2')
    run_cell('logging.info(b)')
    run_cell('logging.info(b)')
    deps = set(compute_slice(4).keys())
    assert deps == {1, 2, 4}, 'got %s' % deps
    slice_size = num_stmts_in_slice(4)
    assert slice_size == 3, 'got %d' % slice_size


@dynamic_only_test
def test_multiple_versions_captured():
    run_cell('x = 0')
    run_cell('logging.info(x); y = 7')
    run_cell('x = 5')
    run_cell('logging.info(x + y)')
    deps = set(compute_slice(4).keys())
    assert deps == {1, 2, 3, 4}, 'got %s' % deps
    slice_size = num_stmts_in_slice(4)
    assert slice_size == 3, 'got %d' % slice_size


def test_version_used_when_live():
    run_cell('x = 0')
    run_cell("""
if True:
    y = 7
else:
    # even though this branch is not taken,
    # liveness-based usage should detect the
    # version of `x` used at the time it was
    # live, meaning cell 1 should get included
    # in the slice
    logging.info(x)
""")
    run_cell('x = 5')
    run_cell('logging.info(x + y)')
    deps = set(compute_slice(4).keys())
    assert deps == {1, 2, 3, 4}, 'got %s' % deps
    slice_size = num_stmts_in_slice(4)
    assert slice_size == 4, 'got %d' % slice_size


def test_no_definitely_spurious_cells():
    run_cell('x = 0')
    run_cell("""
if True:
    y = 7
else:
    # even though this branch is not taken,
    # liveness-based usage should detect the
    # version of `x` used at the time it was
    # live, meaning cell 1 should get included
    # in the slice
    logging.info(x)
""")
    run_cell('x = 5')
    run_cell('logging.info(y)')
    deps = set(compute_slice(4).keys())
    assert deps == {1, 2, 4}, 'got %s' % deps
    slice_size = num_stmts_in_slice(4)
    assert slice_size == 3, 'got %d' % slice_size


def test_parent_usage_includes_child_update():
    run_cell('lst = [3]')
    run_cell('lst[0] += 1')
    run_cell('lst2 = lst + [5]')
    deps = set(compute_slice(3).keys())
    assert deps == {1, 2, 3}, 'got %s' % deps
    slice_size = num_stmts_in_slice(3)
    assert slice_size == 3, 'got %d' % slice_size


def test_object_subscripting():
    run_cell("""
class Foo:
    def __init__(self):
        self.counter = 0
        
    def inc(self):
        self.counter += 1

    def bar(self):
        return {'baz': 0}
""")
    run_cell("obj = Foo()")
    run_cell("name = 'baz'")
    run_cell("obj.bar()[name]")
    run_cell("something_i_dont_care_about = obj.bar()[name]")
    run_cell("something_i_care_about = obj.bar()[name]")
    deps = set(compute_slice(6).keys())
    assert deps == {1, 2, 3, 6}, 'got %s' % deps
    slice_size = num_stmts_in_slice(6)
    assert slice_size == 4, 'got %d' % slice_size


def test_complicated_subscripting():
    run_cell("""
class Foo:
    def __init__(self):
        self.counter = 0
        
    def inc(self):
        self.counter += 1

    def bar(self, indicator):
        if indicator == 1:
            return the_dictionary
        else:
            return x + 5
    
    def new(self, indicator):
        if indicator == 1:
            return Foo()
        
        return Bar()
""")
    run_cell("""
class Bar:
    def __init__(self):
        self.counter = 0
    
    def foo(self, indicator):
        if indicator == 1:
            return the_dictionary
        
        return Foo()
""")
    run_cell("x = 1")
    run_cell("the_dictionary = {'something': 1}")
    run_cell("Foo().new(0).foo(1)")
    deps = set(compute_slice(5).keys())
    assert deps == {1, 2, 4, 5}, 'got %s' % deps
    slice_size = num_stmts_in_slice(5)
    assert slice_size == 4, 'got %d' % slice_size


def test_complicated_subscripting_use_conditional():
    run_cell("""
class Foo:
    def __init__(self):
        self.counter = 0
        
    def inc(self):
        self.counter += 1

    def bar(self, indicator):
        if indicator == 1:
            return the_dictionary
        else:
            return x + 5
    
    def new(self, indicator):
        if indicator == 1:
            return Foo()
        
        return Bar()
""")
    run_cell("""
class Bar:
    def __init__(self):
        self.counter = 0
    
    def foo(self, indicator):
        if indicator == 1:
            return the_dictionary
        
        return Foo()
""")
    run_cell("x = 1")
    run_cell("the_dictionary = {'something': 1}")
    run_cell("""
y = Foo().new(0).foo(1)
z = Foo().new(1).bar(0)
""")
    run_cell("logging.info(z)")
    run_cell("logging.info(y)")
    deps = set(compute_slice(6).keys())
    assert deps == {1, 2, 3, 4, 5, 6}, 'got %s' % deps
    slice_size = num_stmts_in_slice(6)
    assert slice_size == 4, 'got %d' % slice_size
    slice_size = num_stmts_in_slice(7)
    assert slice_size == 5, 'got %d' % slice_size


def test_non_relevant_child_symbol_modified():
    run_cell('lst = [0, 1, 2]')
    run_cell('lst[0] += 1')
    run_cell('lst[0] += 1')
    run_cell('lst[1] += 1')
    run_cell('lst[0] += 1')
    run_cell('lst[0] += 1')
    run_cell('logging.info(lst[1])')
    run_cell('lst[1] = 42')
    deps = set(compute_slice(7).keys())
    assert deps == {1, 4, 7}, 'got %s' % deps
    deps = set(compute_slice(8).keys())
    assert deps == {1, 8}, 'got %s' % deps
    slice_size = num_stmts_in_slice(7)
    assert slice_size == 3, 'got %d' % slice_size


@dynamic_only_test
def test_dynamic_only_increment():
    nbs().mut_settings.static_slicing_enabled = False
    run_cell('x = 0')
    run_cell('x += 1')
    run_cell('logging.info(x)')
    deps = set(compute_slice(3).keys())
    assert deps == {1, 2, 3}, 'got %s' % deps
    slice_size = num_stmts_in_slice(3)
    assert slice_size == 3, 'got %d' % slice_size


@dynamic_only_test
def test_dynamic_only_variable_subscript():
    nbs().mut_settings.static_slicing_enabled = False
    run_cell('lst = [0, 1, 2]')
    run_cell('x = 0')
    run_cell('lst[x] += 1')
    run_cell('lst[x] += 1')
    run_cell('x += 1')
    run_cell('lst[x] += 1')
    run_cell('x -= 1')
    run_cell('lst[x] += 1')
    run_cell('lst[x] += 1')
    run_cell('x += 1')
    run_cell('logging.info(lst[x])')
    deps = set(compute_slice(11).keys())
    assert deps == {1, 2, 5, 6, 7, 10, 11}, 'got %s' % deps
    slice_size = num_stmts_in_slice(11)
    assert slice_size == len(deps), 'got %d' % slice_size


@dynamic_only_test
def test_handler():
    run_cell("""
try:
    r = map()
except:
    success = False""")
    run_cell("""
try:
    logging.info("%d %d", (1, 2))
except:
    success = False""")
    deps = set(compute_slice(2).keys())
    assert deps == {2}, 'got %s' % deps
    slice_size = num_stmts_in_slice(2)
    assert slice_size == 1, 'got %d' % slice_size


@dynamic_only_test
def test_list_delete():
    run_cell('lst = [0, 1, 2, 3, 4, 5, 6]')
    run_cell('lst.append(7)')
    run_cell('del lst[0]')
    run_cell('del lst[2]')
    run_cell('logging.info(lst[1])')
    deps = set(compute_slice(5).keys())
    assert deps == {1, 2, 3, 5}, 'got %s' % deps
    slice_size = num_stmts_in_slice(5)
    assert slice_size == len(deps), 'got %d' % slice_size



# -*- coding: future_annotations -*-
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Set, Tuple


METHODS_WITH_MUTATION_EVEN_FOR_NON_NULL_RETURN: Set[Tuple[int, str]] = set()
METHODS_WITHOUT_MUTATION_EVEN_FOR_NULL_RETURN: Set[Tuple[int, str]] = set()

try:
    import pylab
    pylab_id = id(pylab)
    METHODS_WITH_MUTATION_EVEN_FOR_NON_NULL_RETURN.add((pylab_id, 'figure'))
    METHODS_WITHOUT_MUTATION_EVEN_FOR_NULL_RETURN.add((pylab_id, 'show'))
    METHODS_WITHOUT_MUTATION_EVEN_FOR_NULL_RETURN.add((pylab_id, 'plot'))
except ImportError:
    pass

try:
    import matplotlib.pyplot as plt
    plt_id = id(plt)
    METHODS_WITH_MUTATION_EVEN_FOR_NON_NULL_RETURN.add((plt_id, 'figure'))
    METHODS_WITHOUT_MUTATION_EVEN_FOR_NULL_RETURN.add((plt_id, 'show'))
    METHODS_WITHOUT_MUTATION_EVEN_FOR_NULL_RETURN.add((plt_id, 'plot'))
except ImportError:
    pass

try:
    import d2l.torch as d2l
    d2l_id = id(d2l)
    METHODS_WITHOUT_MUTATION_EVEN_FOR_NULL_RETURN.add((d2l_id, 'plot'))
except ImportError:
    pass



# -*- coding: future_annotations -*-
"""
The idea of this file is to make sure that no debugging "print" statements
make it into production.
"""
import ast
import os

import nbsafety


join = os.path.join
root = join(os.curdir, nbsafety.__name__)


_EXCEPTED_FILES = {
    join(root, '_version.py'),
    join(join(root, 'kernel'), 'install.py'),
}


class ContainsPrintVisitor(ast.NodeVisitor):
    def __init__(self):
        self._found_print_call = False

    def __call__(self, filename: str) -> bool:
        with open(filename, 'r') as f:
            self.visit(ast.parse(f.read()))
        ret = self._found_print_call
        self._found_print_call = False
        return ret

    def visit_Call(self, node: ast.Call):
        self.generic_visit(node)
        if isinstance(node.func, ast.Name) and node.func.id == 'print':
            self._found_print_call = True


def test_no_prints():
    contains_print = ContainsPrintVisitor()
    for path, _, files in os.walk(root):
        for filename in files:
            if not filename.endswith('.py') or filename in _EXCEPTED_FILES:
                continue
            filename = os.path.join(path, filename)
            if filename in _EXCEPTED_FILES:
                continue
            assert not contains_print(filename), f'file {filename} had a print statement!'



# -*- coding: future_annotations -*-
from nbsafety.data_model.annotation_utils import get_type_annotation, make_annotation_string


def _make_annotation_str_for_obj(obj):
    return make_annotation_string(get_type_annotation(obj))


def test_dict():
    empty = _make_annotation_str_for_obj({})
    assert empty == 'Dict', 'got %s' % empty
    assert _make_annotation_str_for_obj({1: 2}) == 'Dict[int, int]'
    assert _make_annotation_str_for_obj({1: 2, 2: 3.0}) == 'Dict[int, float]'
    ann_str = _make_annotation_str_for_obj({1: 2, 2: 3.0, 'foo': 'bar'})
    assert ann_str == 'Dict[Union[int, str], Union[float, str]]', 'got %s' % ann_str
    assert _make_annotation_str_for_obj({1: 2, 2: 3.0, 3: None}) == 'Dict[int, Optional[float]]'


def test_list():
    assert _make_annotation_str_for_obj([]) == 'List'
    assert _make_annotation_str_for_obj([3, 6.0, 9]) == 'List[float]'
    assert _make_annotation_str_for_obj([3, 9]) == 'List[int]'
    ann_str = _make_annotation_str_for_obj([3, None])
    assert ann_str == 'List[Optional[int]]', 'got %s' % ann_str
    assert _make_annotation_str_for_obj([3, None, 42.]) == 'List[Optional[float]]'
    ann_str = _make_annotation_str_for_obj([3, None, 42., 'foo'])
    assert ann_str == 'List[Union[None, float, str]]', 'got %s' % ann_str


def test_set():
    assert _make_annotation_str_for_obj(set()) == 'Set'
    assert _make_annotation_str_for_obj({3, 6.0, 9}) == 'Set[float]'
    assert _make_annotation_str_for_obj({3, 9}) == 'Set[int]'
    assert _make_annotation_str_for_obj({3, None}) == 'Set[Optional[int]]'
    assert _make_annotation_str_for_obj({3, None, 42.}) == 'Set[Optional[float]]'
    assert _make_annotation_str_for_obj({3, None, 42., 'foo'}) == 'Set[Union[None, float, str]]'


def test_tuple():
    assert _make_annotation_str_for_obj(()) == 'Tuple'
    assert _make_annotation_str_for_obj((3, 4)) == 'Tuple[int, int]'
    assert _make_annotation_str_for_obj((3, 4, 5.0)) == 'Tuple[int, int, float]'
    assert _make_annotation_str_for_obj((3, 4, 5, 6)) == 'Tuple[int, ...]'
    assert _make_annotation_str_for_obj((3, 4, 5, 6.0)) == 'Tuple[float, ...]'


class Foo:
    pass


def test_class():
    ann_str = _make_annotation_str_for_obj(Foo)
    assert ann_str == 'Type[{prefix}.Foo]'.format(prefix=__name__), 'got %s' % ann_str
    assert _make_annotation_str_for_obj((Foo, Foo())) == 'Tuple[Type[{prefix}.Foo], {prefix}.Foo]'.format(prefix=__name__)
    assert _make_annotation_str_for_obj([Foo, Foo()]) == 'List[Union[Type[{prefix}.Foo], {prefix}.Foo]]'.format(prefix=__name__)



# -*- coding: future_annotations -*-
import ast

from nbsafety.tracing.ast_eavesdrop import AstEavesdropper
from nbsafety.utils import KeyDict


PROGRAM = """
for i in [foo(x) for x in [1, 2, 3]]:
    try:
        logging.info(i)
    except:
        logging.warning("warning!")
    finally:
        sys.exit(0)
        
if False:
    asdf(qwer(1, 2, a().b[c,d,e](f.g()).h))[a, 7] = foo.bar
"""


def test_ast_rewrite():
    """
    No asserts; just make sure we don't throw an error.
    """
    rewriter = AstEavesdropper(KeyDict())
    assert rewriter.visit(ast.parse(PROGRAM)) is not None
